from notmuch2 import _capi as capi


class NotmuchError(Exception):
    """Base exception for errors originating from the notmuch library.

    Usually this will have two attributes:

    :status: This is a numeric status code corresponding to the error
       code in the notmuch library.  This is normally fairly
       meaningless, it can also often be ``None``.  This exists mostly
       to easily create new errors from notmuch status codes and
       should not normally be used by users.

    :message: A user-facing message for the error.  This can
       occasionally also be ``None``.  Usually you'll want to call
       ``str()`` on the error object instead to get a sensible
       message.
    """

    @classmethod
    def exc_type(cls, status):
        """Return correct exception type for notmuch status."""
        types = {
            capi.lib.NOTMUCH_STATUS_OUT_OF_MEMORY:
                OutOfMemoryError,
            capi.lib.NOTMUCH_STATUS_READ_ONLY_DATABASE:
                ReadOnlyDatabaseError,
            capi.lib.NOTMUCH_STATUS_XAPIAN_EXCEPTION:
                XapianError,
            capi.lib.NOTMUCH_STATUS_FILE_ERROR:
                FileError,
            capi.lib.NOTMUCH_STATUS_FILE_NOT_EMAIL:
                FileNotEmailError,
            capi.lib.NOTMUCH_STATUS_DUPLICATE_MESSAGE_ID:
                DuplicateMessageIdError,
            capi.lib.NOTMUCH_STATUS_NULL_POINTER:
                NullPointerError,
            capi.lib.NOTMUCH_STATUS_TAG_TOO_LONG:
                TagTooLongError,
            capi.lib.NOTMUCH_STATUS_UNBALANCED_FREEZE_THAW:
                UnbalancedFreezeThawError,
            capi.lib.NOTMUCH_STATUS_UNBALANCED_ATOMIC:
                UnbalancedAtomicError,
            capi.lib.NOTMUCH_STATUS_UNSUPPORTED_OPERATION:
                UnsupportedOperationError,
            capi.lib.NOTMUCH_STATUS_UPGRADE_REQUIRED:
                UpgradeRequiredError,
            capi.lib.NOTMUCH_STATUS_PATH_ERROR:
                PathError,
            capi.lib.NOTMUCH_STATUS_ILLEGAL_ARGUMENT:
                IllegalArgumentError,
            capi.lib.NOTMUCH_STATUS_NO_CONFIG:
                NoConfigError,
            capi.lib.NOTMUCH_STATUS_NO_DATABASE:
                NoDatabaseError,
            capi.lib.NOTMUCH_STATUS_DATABASE_EXISTS:
                DatabaseExistsError,
            capi.lib.NOTMUCH_STATUS_BAD_QUERY_SYNTAX:
                QuerySyntaxError,
        }
        return types[status]

    def __new__(cls, *args, **kwargs):
        """Return the correct subclass based on status."""
        # This is simplistic, but the actual __init__ will fail if the
        # signature is wrong anyway.
        if args:
            status = args[0]
        else:
            status = kwargs.get('status', None)
        if status and cls == NotmuchError:
            exc = cls.exc_type(status)
            return exc.__new__(exc, *args, **kwargs)
        else:
            return super().__new__(cls)

    def __init__(self, status=None, message=None):
        self.status = status
        self.message = message

    def __str__(self):
        if self.message:
            return self.message
        elif self.status:
            char_str = capi.lib.notmuch_status_to_string(self.status)
            return capi.ffi.string(char_str).decode(errors='replace')
        else:
            return 'Unknown error'


class OutOfMemoryError(NotmuchError): pass
class ReadOnlyDatabaseError(NotmuchError): pass
class XapianError(NotmuchError): pass
class FileError(NotmuchError): pass
class FileNotEmailError(NotmuchError): pass
class DuplicateMessageIdError(NotmuchError): pass
class NullPointerError(NotmuchError): pass
class TagTooLongError(NotmuchError): pass
class UnbalancedFreezeThawError(NotmuchError): pass
class UnbalancedAtomicError(NotmuchError): pass
class UnsupportedOperationError(NotmuchError): pass
class UpgradeRequiredError(NotmuchError): pass
class PathError(NotmuchError): pass
class IllegalArgumentError(NotmuchError): pass
class NoConfigError(NotmuchError): pass
class NoDatabaseError(NotmuchError): pass
class DatabaseExistsError(NotmuchError): pass
class QuerySyntaxError(NotmuchError): pass

class ObjectDestroyedError(NotmuchError):
    """The object has already been destroyed and it's memory freed.

    This occurs when :meth:`destroy` has been called on the object but
    you still happen to have access to the object.  This should not
    normally occur since you should never call :meth:`destroy` by
    hand.
    """

    def __str__(self):
        if self.message:
            return self.message
        else:
            return 'Memory already freed'



"""The :mod:`notmuch` module provides most of the functionality that a user is
likely to need.

.. note:: The underlying notmuch library is build on a hierarchical
    memory allocator called talloc. All objects derive from a
    top-level :class:`Database` object.

    This means that as soon as an object is deleted, all underlying
    derived objects such as Queries, Messages, Message, and Tags will
    be freed by the underlying library as well. Accessing these
    objects will then lead to segfaults and other unexpected behavior.

    We implement reference counting, so that parent objects can be
    automatically freed when they are not needed anymore. For
    example::

            db = Database('path',create=True)
            msgs = Query(db,'from:myself').search_messages()

    This returns :class:`Messages` which internally contains a
    reference to its parent :class:`Query` object. Otherwise the
    Query() would be immediately freed, taking our *msgs* down with
    it.

    In this case, the above Query() object will be automatically freed
    whenever we delete all derived objects, ie in our case:
    `del(msgs)` would also delete the parent Query. It would not
    delete the parent Database() though, as that is still referenced
    from the variable *db* in which it is stored.

    Pretty much the same is valid for all other objects in the
    hierarchy, such as :class:`Query`, :class:`Messages`,
    :class:`Message`, and :class:`Tags`.
"""

"""
This file is part of notmuch.

Notmuch is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

Notmuch is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with notmuch.  If not, see <https://www.gnu.org/licenses/>.

Copyright 2010-2011 Sebastian Spaeth <Sebastian@SSpaeth.de>
"""
from .database import Database
from .directory import Directory
from .filenames import Filenames
from .message import Message
from .messages import Messages
from .query import Query
from .tag import Tags
from .thread import Thread
from .threads import Threads
from .globals import nmlib
from .errors import (
    STATUS,
    NotmuchError,
    OutOfMemoryError,
    ReadOnlyDatabaseError,
    XapianError,
    FileError,
    FileNotEmailError,
    DuplicateMessageIdError,
    NullPointerError,
    TagTooLongError,
    UnbalancedFreezeThawError,
    UnbalancedAtomicError,
    NotInitializedError,
    UnsupportedOperationError,
    UpgradeRequiredError,
    PathError,
)
from .version import __VERSION__
__LICENSE__ = "GPL v3+"
__AUTHOR__ = 'Sebastian Spaeth <Sebastian@SSpaeth.de>'



"""Pythonic API to the notmuch database.

Creating Objects
================

Only the :class:`Database` object is meant to be created by the user.
All other objects should be created from this initial object.  Users
should consider their signatures implementation details.

Errors
======

All errors occurring due to errors from the underlying notmuch database
are subclasses of the :exc:`NotmuchError`.  Due to memory management
it is possible to try and use an object after it has been freed.  In
this case a :exc:`ObjectDestroyedError` will be raised.

Memory Management
=================

Libnotmuch uses a hierarchical memory allocator, this means all
objects have a strict parent-child relationship and when the parent is
freed all the children are freed as well.  This has some implications
for these Python bindings as parent objects need to be kept alive.
This is normally schielded entirely from the user however and the
Python objects automatically make sure the right references are kept
alive.  It is however the reason the :class:`BaseObject` exists as it
defines the API all Python objects need to implement to work
correctly.

Collections and Containers
==========================

Libnotmuch exposes nearly all collections of things as iterators only.
In these python bindings they have sometimes been exposed as
:class:`collections.abc.Container` instances or subclasses of this
like :class:`collections.abc.Set` or :class:`collections.abc.Mapping`
etc.  This gives a more natural API to work with, e.g. being able to
treat tags as sets.  However it does mean that the
:meth:`__contains__`, :meth:`__len__` and frieds methods on these are
usually more and essentially O(n) rather than O(1) as you might
usually expect from Python containers.
"""

from notmuch2 import _capi
from notmuch2._base import *
from notmuch2._database import *
from notmuch2._errors import *
from notmuch2._message import *
from notmuch2._tags import *
from notmuch2._thread import *


NOTMUCH_TAG_MAX = _capi.lib.NOTMUCH_TAG_MAX
del _capi


# Re-home all the objects to the package.  This leaves __qualname__ intact.
for x in locals().copy().values():
    if hasattr(x, '__module__'):
        x.__module__ = __name__
del x



import pytest

from notmuch2 import _base as base
from notmuch2 import _errors as errors


class TestNotmuchObject:

    def test_no_impl_methods(self):
        class Object(base.NotmuchObject):
            pass
        with pytest.raises(TypeError):
            Object()

    def test_impl_methods(self):

        class Object(base.NotmuchObject):

            def __init__(self):
                pass

            @property
            def alive(self):
                pass

            def _destroy(self, parent=False):
                pass

        Object()

    def test_del(self):
        destroyed = False

        class Object(base.NotmuchObject):

            def __init__(self):
                pass

            @property
            def alive(self):
                pass

            def _destroy(self, parent=False):
                nonlocal destroyed
                destroyed = True

        o = Object()
        o.__del__()
        assert destroyed


class TestMemoryPointer:

    @pytest.fixture
    def obj(self):
        class Cls:
            ptr = base.MemoryPointer()
        return Cls()

    def test_unset(self, obj):
        with pytest.raises(errors.ObjectDestroyedError):
            obj.ptr

    def test_set(self, obj):
        obj.ptr = 'some'
        assert obj.ptr == 'some'

    def test_cleared(self, obj):
        obj.ptr = 'some'
        obj.ptr
        obj.ptr = None
        with pytest.raises(errors.ObjectDestroyedError):
            obj.ptr

    def test_two_instances(self, obj):
        obj2 = obj.__class__()
        obj.ptr = 'foo'
        obj2.ptr = 'bar'
        assert obj.ptr != obj2.ptr


class TestBinString:

    def test_type(self):
        s = base.BinString(b'foo')
        assert isinstance(s, str)

    def test_init_bytes(self):
        s = base.BinString(b'foo')
        assert s == 'foo'

    def test_init_str(self):
        s = base.BinString('foo')
        assert s == 'foo'

    def test_bytes(self):
        s = base.BinString(b'foo')
        assert bytes(s) == b'foo'

    def test_invalid_utf8(self):
        s = base.BinString(b'\x80foo')
        assert s == 'foo'
        assert bytes(s) == b'\x80foo'

    def test_errors(self):
        s = base.BinString(b'\x80foo', errors='replace')
        assert s == '�foo'
        assert bytes(s) == b'\x80foo'

    def test_encoding(self):
        # pound sign: '£' == '\u00a3' latin-1: b'\xa3', utf-8: b'\xc2\xa3'
        with pytest.raises(UnicodeDecodeError):
            base.BinString(b'\xa3', errors='strict')
        s = base.BinString(b'\xa3', encoding='latin-1', errors='strict')
        assert s == '£'
        assert bytes(s) == b'\xa3'



import collections.abc
import weakref

from notmuch2 import _base as base
from notmuch2 import _capi as capi
from notmuch2 import _errors as errors
from notmuch2 import _message as message
from notmuch2 import _tags as tags


__all__ = ['Thread']


class Thread(base.NotmuchObject, collections.abc.Iterable):
    _thread_p = base.MemoryPointer()

    def __init__(self, parent, thread_p, *, db):
        self._parent = parent
        self._thread_p = thread_p
        self._db = db

    @property
    def alive(self):
        if not self._parent.alive:
            return False
        try:
            self._thread_p
        except errors.ObjectDestroyedError:
            return False
        else:
            return True

    def __del__(self):
        self._destroy()

    def _destroy(self):
        if self.alive:
            capi.lib.notmuch_thread_destroy(self._thread_p)
        self._thread_p = None

    @property
    def threadid(self):
        """The thread ID as a :class:`BinString`.

        :raises ObjectDestroyedError: if used after destroyed.
        """
        ret = capi.lib.notmuch_thread_get_thread_id(self._thread_p)
        return base.BinString.from_cffi(ret)

    def __len__(self):
        """Return the number of messages in the thread.

        :raises ObjectDestroyedError: if used after destroyed.
        """
        return capi.lib.notmuch_thread_get_total_messages(self._thread_p)

    def toplevel(self):
        """Return an iterator of the toplevel messages.

        :returns: An iterator yielding :class:`Message` instances.

        :raises ObjectDestroyedError: if used after destroyed.
        """
        msgs_p = capi.lib.notmuch_thread_get_toplevel_messages(self._thread_p)
        return message.MessageIter(self, msgs_p,
                                   db=self._db,
                                   msg_cls=message.OwnedMessage)

    def __iter__(self):
        """Return an iterator over all the messages in the thread.

        :returns: An iterator yielding :class:`Message` instances.

        :raises ObjectDestroyedError: if used after destroyed.
        """
        msgs_p = capi.lib.notmuch_thread_get_messages(self._thread_p)
        return message.MessageIter(self, msgs_p,
                                   db=self._db,
                                   msg_cls=message.OwnedMessage)

    @property
    def matched(self):
        """The number of messages in this thread which matched the query.

        Of the messages in the thread this gives the count of messages
        which did directly match the search query which this thread
        originates from.

        :raises ObjectDestroyedError: if used after destroyed.
        """
        return capi.lib.notmuch_thread_get_matched_messages(self._thread_p)

    @property
    def authors(self):
        """A comma-separated string of all authors in the thread.

        Authors of messages which matched the query the thread was
        retrieved from will be at the head of the string, ordered by
        date of their messages.  Following this will be the authors of
        the other messages in the thread, also ordered by date of
        their messages.  Both groups of authors are separated by the
        ``|`` character.

        :returns: The stringified list of authors.
        :rtype: BinString

        :raises ObjectDestroyedError: if used after destroyed.
        """
        ret = capi.lib.notmuch_thread_get_authors(self._thread_p)
        return base.BinString.from_cffi(ret)

    @property
    def subject(self):
        """The subject of the thread, taken from the first message.

        The thread's subject is taken to be the subject of the first
        message according to query sort order.

        :returns: The thread's subject.
        :rtype: BinString

        :raises ObjectDestroyedError: if used after destroyed.
        """
        ret = capi.lib.notmuch_thread_get_subject(self._thread_p)
        return base.BinString.from_cffi(ret)

    @property
    def first(self):
        """Return the date of the oldest message in the thread.

        The time the first message was sent as an integer number of
        seconds since the *epoch*, 1 Jan 1970.

        :raises ObjectDestroyedError: if used after destroyed.
        """
        return capi.lib.notmuch_thread_get_oldest_date(self._thread_p)

    @property
    def last(self):
        """Return the date of the newest message in the thread.

        The time the last message was sent as an integer number of
        seconds since the *epoch*, 1 Jan 1970.

        :raises ObjectDestroyedError: if used after destroyed.
        """
        return capi.lib.notmuch_thread_get_newest_date(self._thread_p)

    @property
    def tags(self):
        """Return an immutable set with all tags used in this thread.

        This returns an immutable set-like object implementing the
        collections.abc.Set Abstract Base Class.  Due to the
        underlying libnotmuch implementation some operations have
        different performance characteristics then plain set objects.
        Mainly any lookup operation is O(n) rather then O(1).

        Normal usage treats tags as UTF-8 encoded unicode strings so
        they are exposed to Python as normal unicode string objects.
        If you need to handle tags stored in libnotmuch which are not
        valid unicode do check the :class:`ImmutableTagSet` docs for
        how to handle this.

        :rtype: ImmutableTagSet

        :raises ObjectDestroyedError: if used after destroyed.
        """
        try:
            ref = self._cached_tagset
        except AttributeError:
            tagset = None
        else:
            tagset = ref()
        if tagset is None:
            tagset = tags.ImmutableTagSet(
                self, '_thread_p', capi.lib.notmuch_thread_get_tags)
            self._cached_tagset = weakref.ref(tagset)
        return tagset


class ThreadIter(base.NotmuchIter):

    def __init__(self, parent, threads_p, *, db):
        self._db = db
        super().__init__(parent, threads_p,
                         fn_destroy=capi.lib.notmuch_threads_destroy,
                         fn_valid=capi.lib.notmuch_threads_valid,
                         fn_get=capi.lib.notmuch_threads_get,
                         fn_next=capi.lib.notmuch_threads_move_to_next)

    def __next__(self):
        thread_p = super().__next__()
        return Thread(self, thread_p, db=self._db)



#!/usr/bin/env python

from __future__ import absolute_import

try:
  from urllib.parse import quote_plus
  from urllib.parse import unquote_plus
except ImportError:
  from urllib import quote_plus
  from urllib import unquote_plus

from datetime import datetime
from mailbox import MaildirMessage
import mimetypes
import email
import re
import html
import os

import bleach
import web
from notmuch2 import Database
from jinja2 import Environment, FileSystemLoader # FIXME to PackageLoader
from jinja2 import Markup
try:
  import bjoern # from https://github.com/jonashaag/bjoern/
  use_bjoern = True
except:
  use_bjoern = False

# Configuration options
safe_tags = bleach.sanitizer.ALLOWED_TAGS + \
            [u'div', u'span', u'p', u'br', u'table', u'tr', u'td', u'th']
linkify_plaintext = True # delays page load by about 0.02s of 0.20s budget
show_thread_nav = True   # delays page load by about 0.04s of 0.20s budget

prefix = os.environ.get('NMWEB_PREFIX', "http://localhost:8080")
webprefix = os.environ.get('NMWEB_STATIC', prefix + "/static")
cachedir = os.environ.get('NMWEB_CACHE', "static/cache") # special for webpy server; changeable if using your own
cachepath = os.environ.get('NMWEB_CACHE_PATH', cachedir) # location of static cache in the local filesystem

if 'NMWEB_DEBUG' in os.environ:
  web.config.debug = True
else:
  web.config.debug = False

# End of config options

env = Environment(autoescape=True,
                  loader=FileSystemLoader('templates'))

urls = (
  '/', 'index',
  '/search/(.*)', 'search',
  '/show/(.*)', 'show',
)

def urlencode_filter(s):
  if type(s) == 'Markup':
    s = s.unescape()
  s = s.encode('utf8')
  s = quote_plus(s)
  return Markup(s)
env.filters['url'] = urlencode_filter

class index:
  def GET(self):
    web.header('Content-type', 'text/html')
    base = env.get_template('base.html')
    template = env.get_template('index.html')
    db = Database()
    tags = db.tags
    return template.render(tags=tags,
                           title="Notmuch webmail",
                           prefix=prefix,
                           sprefix=webprefix)

class search:
  def GET(self, terms):
    redir = False
    if web.input(terms=None).terms:
      redir = True
      terms = web.input().terms
    terms = unquote_plus (terms)
    if web.input(afters=None).afters:
      afters = web.input(afters=None).afters[:-3]
    else:
      afters = '0'
    if web.input(befores=None).befores:
      befores = web.input(befores=None).befores
    else:
      befores = '4294967296' # 2^32
    try:
      if int(afters) > 0 or int(befores) < 4294967296:
        redir = True
        terms += ' date:@%s..@%s' % (int(afters), int(befores))
    except ValueError:
      pass
    if redir:
      raise web.seeother('/search/%s' % quote_plus(terms.encode('utf8')))
    web.header('Content-type', 'text/html')
    db = Database()
    ts = db.threads(query=terms, sort=Database.SORT.NEWEST_FIRST)
    template = env.get_template('search.html')
    return template.generate(terms=terms,
                             ts=ts,
                             title=terms,
                             prefix=prefix,
                             sprefix=webprefix)

def format_time_range(start, end):
  if end-start < (60*60*24):
    time = datetime.fromtimestamp(start).strftime('%Y %b %d %H:%M')
  else:
    start = datetime.fromtimestamp(start).strftime("%Y %b %d")
    end = datetime.fromtimestamp(end).strftime("%Y %b %d")
    time = "%s through %s" % (start, end)
  return time
env.globals['format_time_range'] = format_time_range

def mailto_addrs(msg,header_name):
  try:
    hdr = msg.header(header_name)
  except LookupError:
    return ''

  frm = email.utils.getaddresses([hdr])
  return ', '.join(['<a href="mailto:%s">%s</a>' % ((l, p) if p else (l, l)) for (p, l) in frm])
env.globals['mailto_addrs'] = mailto_addrs

def link_msg(msg):
  lnk = quote_plus(msg.messageid.encode('utf8'))
  try:
    subj = html.escape(msg.header('Subject'))
  except LookupError:
    subj = ""
  out = '<a href="%s/show/%s">%s</a>' % (prefix, lnk, subj)
  return out
env.globals['link_msg'] = link_msg

def show_msgs(msgs):
  r = '<ul>'
  for msg in msgs:
    red = 'color:black; font-style:normal'
    if msg.matched:
      red = 'color:red; font-style:italic'
    frm = mailto_addrs(msg,'From')
    lnk = link_msg(msg)
    tags = ", ".join(msg.tags)
    rs = show_msgs(msg.replies())
    r += '<li><span style="%s">%s&mdash;%s</span> [%s] %s</li>' % (red, frm, lnk, tags, rs)
  r += '</ul>'
  return r
env.globals['show_msgs'] = show_msgs

# As email.message.walk, but showing close tags as well
def mywalk(self):
  yield self
  if self.is_multipart():
    for subpart in self.get_payload():
      for subsubpart in mywalk(subpart):
        yield subsubpart
    yield 'close-div'

class show:
  def GET(self, mid):
    web.header('Content-type', 'text/html')
    db = Database()
    try:
      m = db.find(mid)
    except:
      raise web.notfound("No such message id.")
    template = env.get_template('show.html')
    # FIXME add reply-all link with email.urils.getaddresses
    # FIXME add forward link using mailto with body parameter?
    return template.render(m=m,
                           mid=mid,
                           title=m.header('Subject'),
                           prefix=prefix,
                           sprefix=webprefix)

def thread_nav(m):
  if not show_thread_nav: return
  db = Database()
  thread = next(db.threads('thread:'+m.threadid))
  prv = None
  found = False
  nxt = None
  for msg in thread:
    if m == msg:
      found = True
    elif not found:
      prv = msg
    else: # found message, but not on this loop
      nxt = msg
      break
  yield "<hr><ul>"
  if prv: yield "<li>Previous message (by thread): %s</li>" % link_msg(prv)
  if nxt: yield "<li>Next message (by thread): %s</li>" % link_msg(nxt)
  yield "</ul><h3>Thread:</h3>"
  # FIXME show now takes three queries instead of 1;
  # can we yield the message body while computing the thread shape?
  thread = next(db.threads('thread:'+m.threadid))
  yield show_msgs(thread.toplevel())
  return
env.globals['thread_nav'] = thread_nav

def format_message(nm_msg, mid):
  fn = list(nm_msg.filenames())[0]
  msg = MaildirMessage(open(fn))
  return format_message_walk(msg, mid)

def decodeAnyway(txt, charset='ascii'):
  try:
    out = txt.decode(charset)
  except:
    try:
      out = txt.decode('utf-8')
    except UnicodeDecodeError:
      out = txt.decode('latin1')
  return out

def require_protocol_prefix(attrs, new=False):
  if not new:
    return attrs
  link_text = attrs[u'_text']
  if link_text.startswith(('http:', 'https:', 'mailto:', 'git:', 'id:')):
    return attrs
  return None

# Bleach doesn't even try to linkify id:... text, so no point invoking this yet
def modify_id_links(attrs, new=False):
  if attrs[(None, u'href')].startswith(u'id:'):
    attrs[(None, u'href')] = prefix + "/show/" + attrs[(None, u'href')][3:]
  return attrs

def css_part_id(content_type, parts=[]):
  c = content_type.replace('/', '-')
  out = "-".join(parts + [c])
  return out

def format_message_walk(msg, mid):
  counter = 0
  cid_refd = []
  parts = ['main']
  for part in mywalk(msg):
    if part == 'close-div':
      parts.pop()
      yield '</div>'
    elif part.get_content_maintype() == 'multipart':
      yield '<div class="multipart-%s" id="%s">' % \
          (part.get_content_subtype(), css_part_id(part.get_content_type(), parts))
      parts.append(part.get_content_subtype())
      if part.get_content_subtype() == 'alternative':
        yield '<ul>'
        for subpart in part.get_payload():
          yield ('<li><a href="#%s">%s</a></li>' %
                 (css_part_id(subpart.get_content_type(), parts),
                  subpart.get_content_type()))
        yield '</ul>'
    elif part.get_content_type() == 'message/rfc822':
      # FIXME extract subject, date, to/cc/from into a separate template and use it here
      yield '<div class="message-rfc822">'
    elif part.get_content_maintype() == 'text':
      if part.get_content_subtype() == 'plain':
        yield '<div id="%s">' % css_part_id(part.get_content_type(), parts)
        yield '<pre>'
        out = part.get_payload(decode=True)
        out = decodeAnyway(out, part.get_content_charset('ascii'))
        out = html.escape(out)
        out = out.encode('ascii', 'xmlcharrefreplace').decode('ascii')
        if linkify_plaintext: out = bleach.linkify(out, callbacks=[require_protocol_prefix])
        yield out
        yield '</pre></div>'
      elif part.get_content_subtype() == 'html':
        yield '<div id="%s">' % css_part_id(part.get_content_type(), parts)
        unb64 = part.get_payload(decode=True)
        decoded = decodeAnyway(unb64, part.get_content_charset('ascii'))
        cid_refd += find_cids(decoded)
        part.set_payload(bleach.clean(replace_cids(decoded, mid), tags=safe_tags).
                         encode(part.get_content_charset('ascii'), 'xmlcharrefreplace'))
        (filename, cid) = link_to_cached_file(part, mid, counter)
        counter += 1
        yield '<iframe class="embedded-html" src="%s"></iframe>' % \
            os.path.join(prefix, cachedir, mid, filename)
        yield '</div>'
      else:
        yield '<div id="%s">' % css_part_id(part.get_content_type(), parts)
        (filename, cid) = link_to_cached_file(part, mid, counter)
        counter += 1
        yield '<a href="%s">%s (%s)</a>' % (os.path.join(prefix,
                                                         cachedir,
                                                         mid,
                                                         filename),
                                            filename,
                                            part.get_content_type())
        yield '</div>'
    elif part.get_content_maintype() == 'image':
      (filename, cid) = link_to_cached_file(part, mid, counter)
      if cid not in cid_refd:
        counter += 1
        yield '<img src="%s" alt="%s">' % (os.path.join(prefix,
                                                        cachedir,
                                                        mid,
                                                        filename),
                                           filename)
    else:
      (filename, cid) = link_to_cached_file(part, mid, counter)
      counter += 1
      yield '<a href="%s">%s (%s)</a>' % (os.path.join(prefix,
                                                       cachedir,
                                                       mid,
                                                       filename),
                                          filename,
                                          part.get_content_type())
env.globals['format_message'] = format_message

def replace_cids(body, mid):
  return body.replace('cid:', os.path.join(prefix, cachedir, mid)+'/')

def find_cids(body):
  return re.findall(r'cid:([^ "\'>]*)', body)

def link_to_cached_file(part, mid, counter):
  filename = part.get_filename()
  if not filename:
    ext = mimetypes.guess_extension(part.get_content_type())
    if not ext:
      ext = '.bin'
    filename = 'part-%03d%s' % (counter, ext)
  try:
    os.makedirs(os.path.join(cachepath, mid))
  except OSError:
    pass
  fn = os.path.join(cachepath, mid, filename) # FIXME escape mid, filename
  fp = open(fn, 'wb')
  if part.get_content_maintype() == 'text':
    data = part.get_payload(decode=True)
    data = decodeAnyway(data, part.get_content_charset('ascii')).encode('utf-8')
  else:
    try:
      data = part.get_payload(decode=True)
    except:
      data = part.get_payload(decode=False)
  if data:
    fp.write(data)
  fp.close()
  if 'Content-ID' in part:
    cid = part['Content-ID']
    if cid[0] == '<' and cid[-1] == '>': cid = cid[1:-1]
    cid_fn = os.path.join(cachepath, mid, cid) # FIXME escape mid, cid
    try:
      os.unlink(cid_fn)
    except OSError:
      pass
    os.link(fn, cid_fn)
    return (filename, cid)
  else:
    return (filename, None)

if __name__ == '__main__':
  app = web.application(urls, globals())
  if use_bjoern:
    bjoern.run(app.wsgifunc(), "127.0.0.1", 8080)
  else:
    app.run()



'''
This file is part of notmuch.

This module handles differences between python2.x and python3.x and
allows the notmuch bindings to support both version families with one
source tree.

Notmuch is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

Notmuch is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with notmuch.  If not, see <https://www.gnu.org/licenses/>.

Copyright 2010 Sebastian Spaeth <Sebastian@SSpaeth.de>
Copyright 2012 Justus Winter <4winter@informatik.uni-hamburg.de>
'''

import sys

if sys.version_info[0] == 2:
    from ConfigParser import SafeConfigParser

    class Python3StringMixIn(object):
        def __str__(self):
            return unicode(self).encode('utf-8')

    def encode_utf8(value):
        '''
        Ensure a nicely utf-8 encoded string to pass to wrapped
        libnotmuch functions.

        C++ code expects strings to be well formatted and unicode
        strings to have no null bytes.
        '''
        if not isinstance(value, basestring):
            raise TypeError('Expected str or unicode, got %s' % type(value))

        if isinstance(value, unicode):
            return value.encode('utf-8', 'replace')

        return value
else:
    from configparser import ConfigParser as SafeConfigParser

    if not hasattr(SafeConfigParser, 'readfp'):   # py >= 3.12
        SafeConfigParser.readfp = SafeConfigParser.read_file

    class Python3StringMixIn(object):
        def __str__(self):
            return self.__unicode__()

    def encode_utf8(value):
        '''
        Ensure a nicely utf-8 encoded string to pass to wrapped
        libnotmuch functions.

        C++ code expects strings to be well formatted and unicode
        strings to have no null bytes.
        '''
        if not isinstance(value, str):
            raise TypeError('Expected str, got %s' % type(value))

        return value.encode('utf-8', 'replace')

# We import the SafeConfigParser class on behalf of other code to cope
# with the differences between Python 2 and 3.
SafeConfigParser # avoid warning about unused import



import setuptools
from _notmuch_config import *

with open(NOTMUCH_VERSION_FILE) as fp:
    VERSION = fp.read().strip()

setuptools.setup(
    name='notmuch2',
    version=VERSION,
    description='Pythonic bindings for the notmuch mail database using CFFI',
    author='Floris Bruynooghe',
    author_email='flub@devork.be',
    setup_requires=['cffi>=1.0.0'],
    install_requires=['cffi>=1.0.0'],
    packages=setuptools.find_packages(exclude=['tests']),
    cffi_modules=['notmuch2/_build.py:ffibuilder'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python :: 3',
        'Topic :: Communications :: Email',
        'Topic :: Software Development :: Libraries',
    ],
)



"""
This file is part of notmuch.

Notmuch is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

Notmuch is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with notmuch.  If not, see <https://www.gnu.org/licenses/>.

Copyright 2010 Sebastian Spaeth <Sebastian@SSpaeth.de>
               Jesse Rosenthal <jrosenthal@jhu.edu>
"""

from .globals import (
    nmlib,
    NotmuchTagsP,
    NotmuchMessageP,
    NotmuchMessagesP,
)
from .errors import (
    NullPointerError,
    NotInitializedError,
)
from .tag import Tags
from .message import Message

class Messages(object):
    r"""Represents a list of notmuch messages

    This object provides an iterator over a list of notmuch messages
    (Technically, it provides a wrapper for the underlying
    *notmuch_messages_t* structure). Do note that the underlying library
    only provides a one-time iterator (it cannot reset the iterator to
    the start). Thus iterating over the function will "exhaust" the list
    of messages, and a subsequent iteration attempt will raise a
    :exc:`NotInitializedError`. If you need to
    re-iterate over a list of messages you will need to retrieve a new
    :class:`Messages` object or cache your :class:`Message`\s in a list
    via::

       msglist = list(msgs)

    You can store and reuse the single :class:`Message` objects as often
    as you want as long as you keep the parent :class:`Messages` object
    around. (Due to hierarchical memory allocation, all derived
    :class:`Message` objects will be invalid when we delete the parent
    :class:`Messages` object, even if it was already exhausted.) So
    this works::

      db   = Database()
      msgs = Query(db,'').search_messages() #get a Messages() object
      msglist = list(msgs)

      # msgs is "exhausted" now and msgs.next() will raise an exception.
      # However it will be kept alive until all retrieved Message()
      # objects are also deleted. If you do e.g. an explicit del(msgs)
      # here, the following lines would fail.

      # You can reiterate over *msglist* however as often as you want.
      # It is simply a list with :class:`Message`s.

      print (msglist[0].get_filename())
      print (msglist[1].get_filename())
      print (msglist[0].get_message_id())


    As :class:`Message` implements both __hash__() and __cmp__(), it is
    possible to make sets out of :class:`Messages` and use set
    arithmetic (this happens in python and will of course be *much*
    slower than redoing a proper query with the appropriate filters::

        s1, s2 = set(msgs1), set(msgs2)
        s.union(s2)
        s1 -= s2
        ...

    Be careful when using set arithmetic between message sets derived
    from different Databases (ie the same database reopened after
    messages have changed). If messages have added or removed associated
    files in the meantime, it is possible that the same message would be
    considered as a different object (as it points to a different file).
    """

    #notmuch_messages_get
    _get = nmlib.notmuch_messages_get
    _get.argtypes = [NotmuchMessagesP]
    _get.restype = NotmuchMessageP

    _collect_tags = nmlib.notmuch_messages_collect_tags
    _collect_tags.argtypes = [NotmuchMessagesP]
    _collect_tags.restype = NotmuchTagsP

    def __init__(self, msgs_p, parent=None):
        """
        :param msgs_p:  A pointer to an underlying *notmuch_messages_t*
             structure. These are not publicly exposed, so a user
             will almost never instantiate a :class:`Messages` object
             herself. They are usually handed back as a result,
             e.g. in :meth:`Query.search_messages`.  *msgs_p* must be
             valid, we will raise an :exc:`NullPointerError` if it is
             `None`.
        :type msgs_p: :class:`ctypes.c_void_p`
        :param parent: The parent object
             (ie :class:`Query`) these tags are derived from. It saves
             a reference to it, so we can automatically delete the db
             object once all derived objects are dead.
        :TODO: Make the iterator work more than once and cache the tags in
               the Python object.(?)
        """
        if not msgs_p:
            raise NullPointerError()

        self._msgs = msgs_p
        #store parent, so we keep them alive as long as self  is alive
        self._parent = parent

    def collect_tags(self):
        """Return the unique :class:`Tags` in the contained messages

        :returns: :class:`Tags`
        :exceptions: :exc:`NotInitializedError` if not init'ed

        .. note::

            :meth:`collect_tags` will iterate over the messages and therefore
            will not allow further iterations.
        """
        if not self._msgs:
            raise NotInitializedError()

        # collect all tags (returns NULL on error)
        tags_p = Messages._collect_tags(self._msgs)
        #reset _msgs as we iterated over it and can do so only once
        self._msgs = None

        if not tags_p:
            raise NullPointerError()
        return Tags(tags_p, self)

    def __iter__(self):
        """ Make Messages an iterator """
        return self

    _valid = nmlib.notmuch_messages_valid
    _valid.argtypes = [NotmuchMessagesP]
    _valid.restype = bool

    _move_to_next = nmlib.notmuch_messages_move_to_next
    _move_to_next.argtypes = [NotmuchMessagesP]
    _move_to_next.restype = None

    def __next__(self):
        if not self._msgs:
            raise NotInitializedError()

        if not self._valid(self._msgs):
            self._msgs = None
            raise StopIteration

        msg = Message(Messages._get(self._msgs), self)
        self._move_to_next(self._msgs)
        return msg
    next = __next__ # python2.x iterator protocol compatibility

    def __nonzero__(self):
        '''
        Implement truth value testing. If __nonzero__ is not
        implemented, the python runtime would fall back to `len(..) >
        0` thus exhausting the iterator.

        :returns: True if the wrapped iterator has at least one more object
                  left.
        '''
        return self._msgs and self._valid(self._msgs)

    _destroy = nmlib.notmuch_messages_destroy
    _destroy.argtypes = [NotmuchMessagesP]
    _destroy.restype = None

    def __del__(self):
        """Close and free the notmuch Messages"""
        if self._msgs:
            self._destroy(self._msgs)

class EmptyMessagesResult(Messages):
    def __init__(self, parent):
        self._msgs = None
        self._parent = parent

    def __next__(self):
        raise StopIteration()
    next = __next__



"""
This file is part of notmuch.

Notmuch is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

Notmuch is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with notmuch.  If not, see <https://www.gnu.org/licenses/>.

Copyright 2010 Sebastian Spaeth <Sebastian@SSpaeth.de>
"""

from ctypes import c_char_p, c_int

from .globals import (
    nmlib,
    Enum,
    Python3StringMixIn,
)

class Status(Enum):
    """Enum with a string representation of a notmuch_status_t value."""
    _status2str = nmlib.notmuch_status_to_string
    _status2str.restype = c_char_p
    _status2str.argtypes = [c_int]

    def __init__(self, statuslist):
        """It is initialized with a list of strings that are available as
        Status().string1 - Status().stringn attributes.
        """
        super(Status, self).__init__(statuslist)

    @classmethod
    def status2str(self, status):
        """Get a (unicode) string representation of a notmuch_status_t value."""
        # define strings for custom error messages
        if status == STATUS.NOT_INITIALIZED:
            return "Operation on uninitialized object impossible."
        return unicode(Status._status2str(status))

STATUS = Status(['SUCCESS',
  'OUT_OF_MEMORY',
  'READ_ONLY_DATABASE',
  'XAPIAN_EXCEPTION',
  'FILE_ERROR',
  'FILE_NOT_EMAIL',
  'DUPLICATE_MESSAGE_ID',
  'NULL_POINTER',
  'TAG_TOO_LONG',
  'UNBALANCED_FREEZE_THAW',
  'UNBALANCED_ATOMIC',
  'UNSUPPORTED_OPERATION',
  'UPGRADE_REQUIRED',
  'PATH_ERROR',
  'NOT_INITIALIZED'])
"""STATUS is a class, whose attributes provide constants that serve as return
indicators for notmuch functions. Currently the following ones are defined. For
possible return values and specific meaning for each method, see the method
description.

  * SUCCESS
  * OUT_OF_MEMORY
  * READ_ONLY_DATABASE
  * XAPIAN_EXCEPTION
  * FILE_ERROR
  * FILE_NOT_EMAIL
  * DUPLICATE_MESSAGE_ID
  * NULL_POINTER
  * TAG_TOO_LONG
  * UNBALANCED_FREEZE_THAW
  * UNBALANCED_ATOMIC
  * UNSUPPORTED_OPERATION
  * UPGRADE_REQUIRED
  * PATH_ERROR
  * NOT_INITIALIZED

Invoke the class method `notmuch.STATUS.status2str` with a status value as
argument to receive a human readable string"""
STATUS.__name__ = 'STATUS'


class NotmuchError(Exception, Python3StringMixIn):
    """Is initiated with a (notmuch.STATUS[, message=None]). It will not
    return an instance of the class NotmuchError, but a derived instance
    of a more specific Error Message, e.g. OutOfMemoryError. Each status
    but SUCCESS has a corresponding subclassed Exception."""

    @classmethod
    def get_exc_subclass(cls, status):
        """Returns a fine grained Exception() type,
        detailing the error status"""
        subclasses = {
            STATUS.OUT_OF_MEMORY: OutOfMemoryError,
            STATUS.READ_ONLY_DATABASE: ReadOnlyDatabaseError,
            STATUS.XAPIAN_EXCEPTION: XapianError,
            STATUS.FILE_ERROR: FileError,
            STATUS.FILE_NOT_EMAIL: FileNotEmailError,
            STATUS.DUPLICATE_MESSAGE_ID: DuplicateMessageIdError,
            STATUS.NULL_POINTER: NullPointerError,
            STATUS.TAG_TOO_LONG: TagTooLongError,
            STATUS.UNBALANCED_FREEZE_THAW: UnbalancedFreezeThawError,
            STATUS.UNBALANCED_ATOMIC: UnbalancedAtomicError,
            STATUS.UNSUPPORTED_OPERATION: UnsupportedOperationError,
            STATUS.UPGRADE_REQUIRED: UpgradeRequiredError,
            STATUS.PATH_ERROR: PathError,
            STATUS.NOT_INITIALIZED: NotInitializedError,
        }
        assert 0 < status <= len(subclasses)
        return subclasses[status]

    def __new__(cls, *args, **kwargs):
        """Return a correct subclass of NotmuchError if needed

        We return a NotmuchError instance if status is None (or 0) and a
        subclass that inherits from NotmuchError depending on the
        'status' parameter otherwise."""
        # get 'status'. Passed in as arg or kwarg?
        status = args[0] if len(args) else kwargs.get('status', None)
        # no 'status' or cls is subclass already, return 'cls' instance
        if not status or cls != NotmuchError:
            return super(NotmuchError, cls).__new__(cls)
        subclass = cls.get_exc_subclass(status)  # which class to use?
        return subclass.__new__(subclass, *args, **kwargs)

    def __init__(self, status=None, message=None):
        self.status = status
        self.message = message

    def __unicode__(self):
        if self.message is not None:
            return self.message
        elif self.status is not None:
            return STATUS.status2str(self.status)
        else:
            return 'Unknown error'


# List of Subclassed exceptions that correspond to STATUS values and are
# subclasses of NotmuchError.
class OutOfMemoryError(NotmuchError):
    status = STATUS.OUT_OF_MEMORY


class ReadOnlyDatabaseError(NotmuchError):
    status = STATUS.READ_ONLY_DATABASE


class XapianError(NotmuchError):
    status = STATUS.XAPIAN_EXCEPTION


class FileError(NotmuchError):
    status = STATUS.FILE_ERROR


class FileNotEmailError(NotmuchError):
    status = STATUS.FILE_NOT_EMAIL


class DuplicateMessageIdError(NotmuchError):
    status = STATUS.DUPLICATE_MESSAGE_ID


class NullPointerError(NotmuchError):
    status = STATUS.NULL_POINTER


class TagTooLongError(NotmuchError):
    status = STATUS.TAG_TOO_LONG


class UnbalancedFreezeThawError(NotmuchError):
    status = STATUS.UNBALANCED_FREEZE_THAW


class UnbalancedAtomicError(NotmuchError):
    status = STATUS.UNBALANCED_ATOMIC


class UnsupportedOperationError(NotmuchError):
    status = STATUS.UNSUPPORTED_OPERATION


class UpgradeRequiredError(NotmuchError):
    status = STATUS.UPGRADE_REQUIRED


class PathError(NotmuchError):
    status = STATUS.PATH_ERROR


class NotInitializedError(NotmuchError):
    """Derived from NotmuchError, this occurs if the underlying data
    structure (e.g. database is not initialized (yet) or an iterator has
    been exhausted. You can test for NotmuchError with .status =
    STATUS.NOT_INITIALIZED"""
    status = STATUS.NOT_INITIALIZED



"""
This file is part of notmuch.

Notmuch is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

Notmuch is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with notmuch.  If not, see <https://www.gnu.org/licenses/>.

Copyright 2010 Sebastian Spaeth <Sebastian@SSpaeth.de>
"""

from ctypes import c_char_p, c_uint, POINTER, byref
from .globals import (
    nmlib,
    Enum,
    _str,
    NotmuchQueryP,
    NotmuchThreadsP,
    NotmuchDatabaseP,
    NotmuchMessagesP,
)
from .errors import (
    NotmuchError,
    NullPointerError,
    NotInitializedError,
)
from .threads import Threads
from .messages import Messages


class Query(object):
    """Represents a search query on an opened :class:`Database`.

    A query selects and filters a subset of messages from the notmuch
    database we derive from.

    :class:`Query` provides an instance attribute :attr:`sort`, which
    contains the sort order (if specified via :meth:`set_sort`) or
    `None`.

    Any function in this class may throw an :exc:`NotInitializedError`
    in case the underlying query object was not set up correctly.

    .. note:: Do remember that as soon as we tear down this object,
           all underlying derived objects such as threads,
           messages, tags etc will be freed by the underlying library
           as well. Accessing these objects will lead to segfaults and
           other unexpected behavior. See above for more details.
    """
    # constants
    SORT = Enum(['OLDEST_FIRST', 'NEWEST_FIRST', 'MESSAGE_ID', 'UNSORTED'])
    """Constants: Sort order in which to return results"""

    def __init__(self, db, querystr):
        """
        :param db: An open database which we derive the Query from.
        :type db: :class:`Database`
        :param querystr: The query string for the message.
        :type querystr: utf-8 encoded str or unicode
        """
        self._db = None
        self._query = None
        self.sort = None
        self.create(db, querystr)

    def _assert_query_is_initialized(self):
        """Raises :exc:`NotInitializedError` if self._query is `None`"""
        if not self._query:
            raise NotInitializedError()

    """notmuch_query_create"""
    _create = nmlib.notmuch_query_create
    _create.argtypes = [NotmuchDatabaseP, c_char_p]
    _create.restype = NotmuchQueryP

    def create(self, db, querystr):
        """Creates a new query derived from a Database

        This function is utilized by __init__() and usually does not need to
        be called directly.

        :param db: Database to create the query from.
        :type db: :class:`Database`
        :param querystr: The query string
        :type querystr: utf-8 encoded str or unicode
        :raises:
            :exc:`NullPointerError` if the query creation failed
                (e.g. too little memory).
            :exc:`NotInitializedError` if the underlying db was not
                initialized.
        """
        db._assert_db_is_initialized()
        # create reference to parent db to keep it alive
        self._db = db
        # create query, return None if too little mem available
        query_p = Query._create(db._db, _str(querystr))
        if not query_p:
            raise NullPointerError
        self._query = query_p

    _set_sort = nmlib.notmuch_query_set_sort
    _set_sort.argtypes = [NotmuchQueryP, c_uint]
    _set_sort.restype = None

    def set_sort(self, sort):
        """Set the sort order future results will be delivered in

        :param sort: Sort order (see :attr:`Query.SORT`)
        """
        self._assert_query_is_initialized()
        self.sort = sort
        self._set_sort(self._query, sort)

    _exclude_tag = nmlib.notmuch_query_add_tag_exclude
    _exclude_tag.argtypes = [NotmuchQueryP, c_char_p]
    _exclude_tag.restype = None

    def exclude_tag(self, tagname):
        """Add a tag that will be excluded from the query results by default.

        This exclusion will be overridden if this tag appears explicitly in the
        query.

        :param tagname: Name of the tag to be excluded
        """
        self._assert_query_is_initialized()
        self._exclude_tag(self._query, _str(tagname))

    """notmuch_query_search_threads"""
    _search_threads = nmlib.notmuch_query_search_threads
    _search_threads.argtypes = [NotmuchQueryP, POINTER(NotmuchThreadsP)]
    _search_threads.restype = c_uint

    def search_threads(self):
        r"""Execute a query for threads

        Execute a query for threads, returning a :class:`Threads` iterator.
        The returned threads are owned by the query and as such, will only be
        valid until the Query is deleted.

        The method sets :attr:`Message.FLAG`\.MATCH for those messages that
        match the query. The method :meth:`Message.get_flag` allows us
        to get the value of this flag.

        :returns: :class:`Threads`
        :raises: :exc:`NullPointerError` if search_threads failed
        """
        self._assert_query_is_initialized()
        threads_p = NotmuchThreadsP() # == NULL
        status = Query._search_threads(self._query, byref(threads_p))
        if status != 0:
            raise NotmuchError(status)

        if not threads_p:
            raise NullPointerError
        return Threads(threads_p, self)

    """notmuch_query_search_messages_st"""
    _search_messages = nmlib.notmuch_query_search_messages
    _search_messages.argtypes = [NotmuchQueryP, POINTER(NotmuchMessagesP)]
    _search_messages.restype = c_uint

    def search_messages(self):
        """Filter messages according to the query and return
        :class:`Messages` in the defined sort order

        :returns: :class:`Messages`
        :raises: :exc:`NullPointerError` if search_messages failed
        """
        self._assert_query_is_initialized()
        msgs_p = NotmuchMessagesP() # == NULL
        status = Query._search_messages(self._query, byref(msgs_p))
        if status != 0:
            raise NotmuchError(status)

        if not msgs_p:
            raise NullPointerError
        return Messages(msgs_p, self)

    _count_messages = nmlib.notmuch_query_count_messages
    _count_messages.argtypes = [NotmuchQueryP, POINTER(c_uint)]
    _count_messages.restype = c_uint

    def count_messages(self):
        '''
        This function performs a search and returns Xapian's best
        guess as to the number of matching messages.

        :returns: the estimated number of messages matching this query
        :rtype:   int
        '''
        self._assert_query_is_initialized()
        count = c_uint(0)
        status = Query._count_messages(self._query, byref(count))
        if status != 0:
            raise NotmuchError(status)
        return count.value

    _count_threads = nmlib.notmuch_query_count_threads
    _count_threads.argtypes = [NotmuchQueryP, POINTER(c_uint)]
    _count_threads.restype = c_uint

    def count_threads(self):
        '''
        This function performs a search and returns the number of
        unique thread IDs in the matching messages. This is the same
        as number of threads matching a search.

        Note that this is a significantly heavier operation than
        meth:`Query.count_messages`.

        :returns: the number of threads returned by this query
        :rtype:   int
        '''
        self._assert_query_is_initialized()
        count = c_uint(0)
        status = Query._count_threads(self._query, byref(count))
        if status != 0:
            raise NotmuchError(status)
        return count.value

    _destroy = nmlib.notmuch_query_destroy
    _destroy.argtypes = [NotmuchQueryP]
    _destroy.restype = None

    def __del__(self):
        """Close and free the Query"""
        if self._query:
            self._destroy(self._query)



"""
This file is part of notmuch.

Notmuch is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

Notmuch is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with notmuch.  If not, see <https://www.gnu.org/licenses/>.

Copyright 2010 Sebastian Spaeth <Sebastian@SSpaeth.de>
"""

import os
import codecs
import warnings
from ctypes import c_char_p, c_void_p, c_uint, byref, POINTER
from .compat import SafeConfigParser
from .globals import (
    nmlib,
    Enum,
    _str,
    NotmuchConfigListP,
    NotmuchDatabaseP,
    NotmuchDirectoryP,
    NotmuchIndexoptsP,
    NotmuchMessageP,
    NotmuchTagsP,
)
from .errors import (
    STATUS,
    FileError,
    NotmuchError,
    NullPointerError,
    NotInitializedError,
)
from .message import Message
from .tag import Tags
from .query import Query
from .directory import Directory

class Database(object):
    """The :class:`Database` is the highest-level object that notmuch
    provides. It references a notmuch database, and can be opened in
    read-only or read-write mode. A :class:`Query` can be derived from
    or be applied to a specific database to find messages. Also adding
    and removing messages to the database happens via this
    object. Modifications to the database are not atmic by default (see
    :meth:`begin_atomic`) and once a database has been modified, all
    other database objects pointing to the same data-base will throw an
    :exc:`XapianError` as the underlying database has been
    modified. Close and reopen the database to continue working with it.

    :class:`Database` objects implement the context manager protocol
    so you can use the :keyword:`with` statement to ensure that the
    database is properly closed. See :meth:`close` for more
    information.

    .. note::

        Any function in this class can and will throw an
        :exc:`NotInitializedError` if the database was not initialized
        properly.
    """
    _std_db_path = None
    """Class attribute to cache user's default database"""

    MODE = Enum(['READ_ONLY', 'READ_WRITE'])
    """Constants: Mode in which to open the database"""

    DECRYPTION_POLICY = Enum(['FALSE', 'TRUE', 'AUTO', 'NOSTASH'])
    """Constants: policies for decrypting messages during indexing"""

    """notmuch_database_get_directory"""
    _get_directory = nmlib.notmuch_database_get_directory
    _get_directory.argtypes = [NotmuchDatabaseP, c_char_p, POINTER(NotmuchDirectoryP)]
    _get_directory.restype = c_uint

    """notmuch_database_get_path"""
    _get_path = nmlib.notmuch_database_get_path
    _get_path.argtypes = [NotmuchDatabaseP]
    _get_path.restype = c_char_p

    """notmuch_database_get_version"""
    _get_version = nmlib.notmuch_database_get_version
    _get_version.argtypes = [NotmuchDatabaseP]
    _get_version.restype = c_uint

    """notmuch_database_get_revision"""
    _get_revision = nmlib.notmuch_database_get_revision
    _get_revision.argtypes = [NotmuchDatabaseP, POINTER(c_char_p)]
    _get_revision.restype = c_uint

    """notmuch_database_open"""
    _open = nmlib.notmuch_database_open
    _open.argtypes = [c_char_p, c_uint, POINTER(NotmuchDatabaseP)]
    _open.restype = c_uint

    """notmuch_database_upgrade"""
    _upgrade = nmlib.notmuch_database_upgrade
    _upgrade.argtypes = [NotmuchDatabaseP, c_void_p, c_void_p]
    _upgrade.restype = c_uint

    """ notmuch_database_find_message"""
    _find_message = nmlib.notmuch_database_find_message
    _find_message.argtypes = [NotmuchDatabaseP, c_char_p,
                              POINTER(NotmuchMessageP)]
    _find_message.restype = c_uint

    """notmuch_database_find_message_by_filename"""
    _find_message_by_filename = nmlib.notmuch_database_find_message_by_filename
    _find_message_by_filename.argtypes = [NotmuchDatabaseP, c_char_p,
                                          POINTER(NotmuchMessageP)]
    _find_message_by_filename.restype = c_uint

    """notmuch_database_get_all_tags"""
    _get_all_tags = nmlib.notmuch_database_get_all_tags
    _get_all_tags.argtypes = [NotmuchDatabaseP]
    _get_all_tags.restype = NotmuchTagsP

    """notmuch_database_create"""
    _create = nmlib.notmuch_database_create
    _create.argtypes = [c_char_p, POINTER(NotmuchDatabaseP)]
    _create.restype = c_uint

    def __init__(self, path = None, create = False,
                 mode = MODE.READ_ONLY):
        """If *path* is `None`, we will try to read a users notmuch
        configuration and use his configured database. The location of the
        configuration file can be specified through the environment variable
        *NOTMUCH_CONFIG*, falling back to the default `~/.notmuch-config`.

        If *create* is `True`, the database will always be created in
        :attr:`MODE`.READ_WRITE mode. Default mode for opening is READ_ONLY.

        :param path:   Directory to open/create the database in (see
                       above for behavior if `None`)
        :type path:    `str` or `None`
        :param create: Pass `False` to open an existing, `True` to create a new
                       database.
        :type create:  bool
        :param mode:   Mode to open a database in. Is always
                       :attr:`MODE`.READ_WRITE when creating a new one.
        :type mode:    :attr:`MODE`
        :raises: :exc:`NotmuchError` or derived exception in case of
            failure.
        """
        self._db = None
        self.mode = mode
        if path is None:
            # no path specified. use a user's default database
            if Database._std_db_path is None:
                #the following line throws a NotmuchError if it fails
                Database._std_db_path = self._get_user_default_db()
            path = Database._std_db_path

        if create == False:
            self.open(path, mode)
        else:
            self.create(path)

    _destroy = nmlib.notmuch_database_destroy
    _destroy.argtypes = [NotmuchDatabaseP]
    _destroy.restype = c_uint

    def __del__(self):
        if self._db:
            status = self._destroy(self._db)
            if status != STATUS.SUCCESS:
                raise NotmuchError(status)

    def _assert_db_is_initialized(self):
        """Raises :exc:`NotInitializedError` if self._db is `None`"""
        if not self._db:
            raise NotInitializedError()

    def create(self, path):
        """Creates a new notmuch database

        This function is used by __init__() and usually does not need
        to be called directly. It wraps the underlying
        *notmuch_database_create* function and creates a new notmuch
        database at *path*. It will always return a database in :attr:`MODE`
        .READ_WRITE mode as creating an empty database for
        reading only does not make a great deal of sense.

        :param path: A directory in which we should create the database.
        :type path: str
        :raises: :exc:`NotmuchError` in case of any failure
                    (possibly after printing an error message on stderr).
        """
        if self._db:
            raise NotmuchError(message="Cannot create db, this Database() "
                                       "already has an open one.")

        db = NotmuchDatabaseP()
        status = Database._create(_str(path), byref(db))

        if status != STATUS.SUCCESS:
            raise NotmuchError(status)
        self._db = db
        return status

    def open(self, path, mode=0):
        """Opens an existing database

        This function is used by __init__() and usually does not need
        to be called directly. It wraps the underlying
        *notmuch_database_open* function.

        :param status: Open the database in read-only or read-write mode
        :type status:  :attr:`MODE`
        :raises: Raises :exc:`NotmuchError` in case of any failure
                    (possibly after printing an error message on stderr).
        """
        db = NotmuchDatabaseP()
        status = Database._open(_str(path), mode, byref(db))

        if status != STATUS.SUCCESS:
            raise NotmuchError(status)
        self._db = db
        return status

    _close = nmlib.notmuch_database_close
    _close.argtypes = [NotmuchDatabaseP]
    _close.restype = c_uint

    def close(self):
        '''
        Closes the notmuch database.

        .. warning::

            This function closes the notmuch database. From that point
            on every method invoked on any object ever derived from
            the closed database may cease to function and raise a
            NotmuchError.
        '''
        if self._db:
            status = self._close(self._db)
            if status != STATUS.SUCCESS:
                raise NotmuchError(status)

    def __enter__(self):
        '''
        Implements the context manager protocol.
        '''
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        '''
        Implements the context manager protocol.
        '''
        self.close()

    def get_path(self):
        """Returns the file path of an open database"""
        self._assert_db_is_initialized()
        return Database._get_path(self._db).decode('utf-8')

    def get_version(self):
        """Returns the database format version

        :returns: The database version as positive integer
        """
        self._assert_db_is_initialized()
        return Database._get_version(self._db)

    def get_revision (self):
        """Returns the committed database revision and UUID

        :returns: (revision, uuid) The database revision as a positive integer
        and the UUID of the database.
        """
        self._assert_db_is_initialized()
        uuid = c_char_p ()
        revision = Database._get_revision(self._db, byref (uuid))
        return (revision, uuid.value.decode ('utf-8'))

    _needs_upgrade = nmlib.notmuch_database_needs_upgrade
    _needs_upgrade.argtypes = [NotmuchDatabaseP]
    _needs_upgrade.restype = bool

    def needs_upgrade(self):
        """Does this database need to be upgraded before writing to it?

        If this function returns `True` then no functions that modify the
        database (:meth:`index_file`,
        :meth:`Message.add_tag`, :meth:`Directory.set_mtime`,
        etc.) will work unless :meth:`upgrade` is called successfully first.

        :returns: `True` or `False`
        """
        self._assert_db_is_initialized()
        return self._needs_upgrade(self._db)

    def upgrade(self):
        """Upgrades the current database

        After opening a database in read-write mode, the client should
        check if an upgrade is needed (notmuch_database_needs_upgrade) and
        if so, upgrade with this function before making any modifications.

        NOT IMPLEMENTED: The optional progress_notify callback can be
        used by the caller to provide progress indication to the
        user. If non-NULL it will be called periodically with
        'progress' as a floating-point value in the range of [0.0..1.0]
        indicating the progress made so far in the upgrade process.

        :TODO: catch exceptions, document return values and etc...
        """
        self._assert_db_is_initialized()
        status = Database._upgrade(self._db, None, None)
        # TODO: catch exceptions, document return values and etc
        return status

    _begin_atomic = nmlib.notmuch_database_begin_atomic
    _begin_atomic.argtypes = [NotmuchDatabaseP]
    _begin_atomic.restype = c_uint

    def begin_atomic(self):
        """Begin an atomic database operation

        Any modifications performed between a successful
        :meth:`begin_atomic` and a :meth:`end_atomic` will be applied to
        the database atomically.  Note that, unlike a typical database
        transaction, this only ensures atomicity, not durability;
        neither begin nor end necessarily flush modifications to disk.

        :returns: :attr:`STATUS`.SUCCESS or raises
        :raises: :exc:`NotmuchError`: :attr:`STATUS`.XAPIAN_EXCEPTION
                    Xapian exception occurred; atomic section not entered.

        *Added in notmuch 0.9*"""
        self._assert_db_is_initialized()
        status = self._begin_atomic(self._db)
        if status != STATUS.SUCCESS:
            raise NotmuchError(status)
        return status

    _end_atomic = nmlib.notmuch_database_end_atomic
    _end_atomic.argtypes = [NotmuchDatabaseP]
    _end_atomic.restype = c_uint

    def end_atomic(self):
        """Indicate the end of an atomic database operation

        See :meth:`begin_atomic` for details.

        :returns: :attr:`STATUS`.SUCCESS or raises

        :raises:
            :exc:`NotmuchError`:
                :attr:`STATUS`.XAPIAN_EXCEPTION
                    A Xapian exception occurred; atomic section not
                    ended.
                :attr:`STATUS`.UNBALANCED_ATOMIC:
                    end_atomic has been called more times than begin_atomic.

        *Added in notmuch 0.9*"""
        self._assert_db_is_initialized()
        status = self._end_atomic(self._db)
        if status != STATUS.SUCCESS:
            raise NotmuchError(status)
        return status

    def get_directory(self, path):
        """Returns a :class:`Directory` of path,

        :param path: An unicode string containing the path relative to the path
              of database (see :meth:`get_path`), or else should be an absolute
              path with initial components that match the path of 'database'.
        :returns: :class:`Directory` or raises an exception.
        :raises: :exc:`FileError` if path is not relative database or absolute
                 with initial components same as database.
        """
        self._assert_db_is_initialized()

        # sanity checking if path is valid, and make path absolute
        if path and path[0] == os.sep:
            # we got an absolute path
            if not path.startswith(self.get_path()):
                # but its initial components are not equal to the db path
                raise FileError('Database().get_directory() called '
                                'with a wrong absolute path')
            abs_dirpath = path
        else:
            #we got a relative path, make it absolute
            abs_dirpath = os.path.abspath(os.path.join(self.get_path(), path))

        dir_p = NotmuchDirectoryP()
        status = Database._get_directory(self._db, _str(path), byref(dir_p))

        if status != STATUS.SUCCESS:
            raise NotmuchError(status)
        if not dir_p:
            return None

        # return the Directory, init it with the absolute path
        return Directory(abs_dirpath, dir_p, self)

    _get_default_indexopts = nmlib.notmuch_database_get_default_indexopts
    _get_default_indexopts.argtypes = [NotmuchDatabaseP]
    _get_default_indexopts.restype = NotmuchIndexoptsP

    _indexopts_set_decrypt_policy = nmlib.notmuch_indexopts_set_decrypt_policy
    _indexopts_set_decrypt_policy.argtypes = [NotmuchIndexoptsP, c_uint]
    _indexopts_set_decrypt_policy.restype = None

    _indexopts_destroy = nmlib.notmuch_indexopts_destroy
    _indexopts_destroy.argtypes = [NotmuchIndexoptsP]
    _indexopts_destroy.restype = None

    _index_file = nmlib.notmuch_database_index_file
    _index_file.argtypes = [NotmuchDatabaseP, c_char_p,
                             c_void_p,
                             POINTER(NotmuchMessageP)]
    _index_file.restype = c_uint

    def index_file(self, filename, sync_maildir_flags=False, decrypt_policy=None):
        """Adds a new message to the database

        :param filename: should be a path relative to the path of the
            open database (see :meth:`get_path`), or else should be an
            absolute filename with initial components that match the
            path of the database.

            The file should be a single mail message (not a
            multi-message mbox) that is expected to remain at its
            current location, since the notmuch database will reference
            the filename, and will not copy the entire contents of the
            file.

        :param sync_maildir_flags: If the message contains Maildir
            flags, we will -depending on the notmuch configuration- sync
            those tags to initial notmuch tags, if set to `True`. It is
            `False` by default to remain consistent with the libnotmuch
            API. You might want to look into the underlying method
            :meth:`Message.maildir_flags_to_tags`.

        :param decrypt_policy: If the message contains any encrypted
            parts, and decrypt_policy is set to
            :attr:`DECRYPTION_POLICY`.TRUE, notmuch will try to
            decrypt the message and index the cleartext, stashing any
            discovered session keys.  If it is set to
            :attr:`DECRYPTION_POLICY`.FALSE, it will never try to
            decrypt during indexing.  If it is set to
            :attr:`DECRYPTION_POLICY`.AUTO, then it will try to use
            any stashed session keys it knows about, but will not try
            to access the user's secret keys.
            :attr:`DECRYPTION_POLICY`.NOSTASH behaves the same as
            :attr:`DECRYPTION_POLICY`.TRUE except that no session keys
            are stashed in the database.  If decrypt_policy is set to
            None (the default), then the database itself will decide
            whether to decrypt, based on the `index.decrypt`
            configuration setting (see notmuch-config(1)).

        :returns: On success, we return

           1) a :class:`Message` object that can be used for things
              such as adding tags to the just-added message.
           2) one of the following :attr:`STATUS` values:

              :attr:`STATUS`.SUCCESS
                  Message successfully added to database.
              :attr:`STATUS`.DUPLICATE_MESSAGE_ID
                  Message has the same message ID as another message already
                  in the database. The new filename was successfully added
                  to the list of the filenames for the existing message.

        :rtype:   2-tuple(:class:`Message`, :attr:`STATUS`)

        :raises: Raises a :exc:`NotmuchError` with the following meaning.
              If such an exception occurs, nothing was added to the database.

              :attr:`STATUS`.FILE_ERROR
                      An error occurred trying to open the file, (such as
                      permission denied, or file not found, etc.).
              :attr:`STATUS`.FILE_NOT_EMAIL
                      The contents of filename don't look like an email
                      message.
              :attr:`STATUS`.READ_ONLY_DATABASE
                      Database was opened in read-only mode so no message can
                      be added.
        """
        self._assert_db_is_initialized()
        msg_p = NotmuchMessageP()
        indexopts = c_void_p(None)
        if decrypt_policy is not None:
            indexopts = self._get_default_indexopts(self._db)
            self._indexopts_set_decrypt_policy(indexopts, decrypt_policy)

        status = self._index_file(self._db, _str(filename), indexopts, byref(msg_p))

        if indexopts:
            self._indexopts_destroy(indexopts)

        if not status in [STATUS.SUCCESS, STATUS.DUPLICATE_MESSAGE_ID]:
            raise NotmuchError(status)

        #construct Message() and return
        msg = Message(msg_p, self)
        #automatic sync initial tags from Maildir flags
        if sync_maildir_flags:
            msg.maildir_flags_to_tags()
        return (msg, status)

    def add_message(self, filename, sync_maildir_flags=False):
        """Deprecated alias for :meth:`index_file`
        """
        warnings.warn(
                "This function is deprecated and will be removed in the future, use index_file.", DeprecationWarning)

        return self.index_file(filename, sync_maildir_flags=sync_maildir_flags)

    _remove_message = nmlib.notmuch_database_remove_message
    _remove_message.argtypes = [NotmuchDatabaseP, c_char_p]
    _remove_message.restype = c_uint

    def remove_message(self, filename):
        """Removes a message (filename) from the given notmuch database

        Note that only this particular filename association is removed from
        the database. If the same message (as determined by the message ID)
        is still available via other filenames, then the message will
        persist in the database for those filenames. When the last filename
        is removed for a particular message, the database content for that
        message will be entirely removed.

        :returns: A :attr:`STATUS` value with the following meaning:

             :attr:`STATUS`.SUCCESS
               The last filename was removed and the message was removed
               from the database.
             :attr:`STATUS`.DUPLICATE_MESSAGE_ID
               This filename was removed but the message persists in the
               database with at least one other filename.

        :raises: Raises a :exc:`NotmuchError` with the following meaning.
             If such an exception occurs, nothing was removed from the
             database.

             :attr:`STATUS`.READ_ONLY_DATABASE
               Database was opened in read-only mode so no message can be
               removed.
        """
        self._assert_db_is_initialized()
        status = self._remove_message(self._db, _str(filename))
        if status not in [STATUS.SUCCESS, STATUS.DUPLICATE_MESSAGE_ID]:
            raise NotmuchError(status)
        return status

    def find_message(self, msgid):
        """Returns a :class:`Message` as identified by its message ID

        Wraps the underlying *notmuch_database_find_message* function.

        :param msgid: The message ID
        :type msgid: unicode or str
        :returns: :class:`Message` or `None` if no message is found.
        :raises:
            :exc:`OutOfMemoryError`
                  If an Out-of-memory occurred while constructing the message.
            :exc:`XapianError`
                  In case of a Xapian Exception. These exceptions
                  include "Database modified" situations, e.g. when the
                  notmuch database has been modified by another program
                  in the meantime. In this case, you should close and
                  reopen the database and retry.
            :exc:`NotInitializedError` if
                    the database was not initialized.
        """
        self._assert_db_is_initialized()
        msg_p = NotmuchMessageP()
        status = Database._find_message(self._db, _str(msgid), byref(msg_p))
        if status != STATUS.SUCCESS:
            raise NotmuchError(status)
        return msg_p and Message(msg_p, self) or None

    def find_message_by_filename(self, filename):
        """Find a message with the given filename

        :returns: If the database contains a message with the given
            filename, then a class:`Message:` is returned.  This
            function returns None if no message is found with the given
            filename.

        :raises: :exc:`OutOfMemoryError` if an Out-of-memory occurred while
                 constructing the message.
        :raises: :exc:`XapianError` in case of a Xapian Exception.
                 These exceptions include "Database modified"
                 situations, e.g. when the notmuch database has been
                 modified by another program in the meantime. In this
                 case, you should close and reopen the database and
                 retry.
        :raises: :exc:`NotInitializedError` if the database was not
                 initialized.

        *Added in notmuch 0.9*"""
        self._assert_db_is_initialized()

        msg_p = NotmuchMessageP()
        status = Database._find_message_by_filename(self._db, _str(filename),
                                                    byref(msg_p))
        if status != STATUS.SUCCESS:
            raise NotmuchError(status)
        return msg_p and Message(msg_p, self) or None

    def get_all_tags(self):
        """Returns :class:`Tags` with a list of all tags found in the database

        :returns: :class:`Tags`
        :exception: :exc:`NotmuchError` with :attr:`STATUS`.NULL_POINTER
                    on error
        """
        self._assert_db_is_initialized()
        tags_p = Database._get_all_tags(self._db)
        if not tags_p:
            raise NullPointerError()
        return Tags(tags_p, self)

    def create_query(self, querystring):
        """Returns a :class:`Query` derived from this database

        This is a shorthand method for doing::

          # short version
          # Automatically frees the Database() when 'q' is deleted

          q  = Database(dbpath).create_query('from:"Biene Maja"')

          # long version, which is functionally equivalent but will keep the
          # Database in the 'db' variable around after we delete 'q':

          db = Database(dbpath)
          q  = Query(db,'from:"Biene Maja"')

        This function is a python extension and not in the underlying C API.
        """
        return Query(self, querystring)

    """notmuch_database_status_string"""
    _status_string = nmlib.notmuch_database_status_string
    _status_string.argtypes = [NotmuchDatabaseP]
    _status_string.restype = c_char_p

    def status_string(self):
        """Returns the status string of the database

        This is sometimes used for additional error reporting
        """
        self._assert_db_is_initialized()
        s = Database._status_string(self._db)
        if s:
            return s.decode('utf-8', 'ignore')
        return s

    def __repr__(self):
        return "'Notmuch DB " + self.get_path() + "'"

    def _get_user_default_db(self):
        """ Reads a user's notmuch config and returns his db location

        Throws a NotmuchError if it cannot find it"""
        config = SafeConfigParser()
        conf_f = os.getenv('NOTMUCH_CONFIG',
                           os.path.expanduser('~/.notmuch-config'))
        config.readfp(codecs.open(conf_f, 'r', 'utf-8'))
        if not config.has_option('database', 'path'):
            raise NotmuchError(message="No DB path specified"
                                       " and no user default found")
        db_path = config.get('database', 'path')
        if not os.path.isabs(db_path):
            db_path = os.path.expanduser(os.path.join("~", db_path))
        return db_path

    """notmuch_database_get_config"""
    _get_config = nmlib.notmuch_database_get_config
    _get_config.argtypes = [NotmuchDatabaseP, c_char_p, POINTER(c_char_p)]
    _get_config.restype = c_uint

    def get_config(self, key):
        """Return the value of the given config key.

        Note that only config values that are stored in the database are
        searched and returned.  The config file is not read.

        :param key: the config key under which a value should be looked up, it
                    should probably be in the form "section.key"
        :type key:  str
        :returns:   the config value or the empty string if no value is present
                    for that key
        :rtype:     str
        :raises:    :exc:`NotmuchError` in case of failure.

        """
        self._assert_db_is_initialized()
        return_string = c_char_p()
        status = self._get_config(self._db, _str(key), byref(return_string))
        if status != STATUS.SUCCESS:
            raise NotmuchError(status)
        return return_string.value.decode('utf-8')

    """notmuch_database_get_config_list"""
    _get_config_list = nmlib.notmuch_database_get_config_list
    _get_config_list.argtypes = [NotmuchDatabaseP, c_char_p,
                                 POINTER(NotmuchConfigListP)]
    _get_config_list.restype = c_uint

    _config_list_valid = nmlib.notmuch_config_list_valid
    _config_list_valid.argtypes = [NotmuchConfigListP]
    _config_list_valid.restype = bool

    _config_list_key = nmlib.notmuch_config_list_key
    _config_list_key.argtypes = [NotmuchConfigListP]
    _config_list_key.restype = c_char_p

    _config_list_value = nmlib.notmuch_config_list_value
    _config_list_value.argtypes = [NotmuchConfigListP]
    _config_list_value.restype = c_char_p

    _config_list_move_to_next = nmlib.notmuch_config_list_move_to_next
    _config_list_move_to_next.argtypes = [NotmuchConfigListP]
    _config_list_move_to_next.restype = None

    _config_list_destroy = nmlib.notmuch_config_list_destroy
    _config_list_destroy.argtypes = [NotmuchConfigListP]
    _config_list_destroy.restype = None

    def get_configs(self, prefix=''):
        """Return a generator of key, value pairs where the start of key
        matches the given prefix

        Note that only config values that are stored in the database are
        searched and returned.  The config file is not read.  If no `prefix` is
        given all config values are returned.

        This could be used to get all named queries into a dict for example::

            queries = {k[6:]: v for k, v in db.get_configs('query.')}

        :param prefix: a string by which the keys should be selected
        :type prefix:  str
        :yields:       all key-value pairs where `prefix` matches the beginning
                       of the key
        :ytype:        pairs of str
        :raises:      :exc:`NotmuchError` in case of failure.

        """
        self._assert_db_is_initialized()
        config_list_p = NotmuchConfigListP()
        status = self._get_config_list(self._db, _str(prefix),
                                       byref(config_list_p))
        if status != STATUS.SUCCESS:
            raise NotmuchError(status)
        while self._config_list_valid(config_list_p):
            key = self._config_list_key(config_list_p).decode('utf-8')
            value = self._config_list_value(config_list_p).decode('utf-8')
            yield key, value
            self._config_list_move_to_next(config_list_p)

    """notmuch_database_set_config"""
    _set_config = nmlib.notmuch_database_set_config
    _set_config.argtypes = [NotmuchDatabaseP, c_char_p, c_char_p]
    _set_config.restype = c_uint

    def set_config(self, key, value):
        """Set a config value in the notmuch database.

        If an empty string is provided as `value` the `key` is unset!

        :param key:   the key to set
        :type key:    str
        :param value: the value to store under `key`
        :type value:  str
        :returns:     None
        :raises:      :exc:`NotmuchError` in case of failure.

        """
        self._assert_db_is_initialized()
        status = self._set_config(self._db, _str(key), _str(value))
        if status != STATUS.SUCCESS:
            raise NotmuchError(status)



"""
This file is part of notmuch.

Notmuch is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

Notmuch is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with notmuch.  If not, see <https://www.gnu.org/licenses/>.

Copyright 2010 Sebastian Spaeth <Sebastian@SSpaeth.de>
"""
from ctypes import c_char_p
from .globals import (
    nmlib,
    Python3StringMixIn,
    NotmuchTagsP,
)
from .errors import (
    NullPointerError,
    NotInitializedError,
)


class Tags(Python3StringMixIn):
    """Represents a list of notmuch tags

    This object provides an iterator over a list of notmuch tags (which
    are unicode instances).

    Do note that the underlying library only provides a one-time
    iterator (it cannot reset the iterator to the start). Thus iterating
    over the function will "exhaust" the list of tags, and a subsequent
    iteration attempt will raise a :exc:`NotInitializedError`.
    Also note, that any function that uses iteration (nearly all) will
    also exhaust the tags. So both::

      for tag in tags: print tag

    as well as::

       number_of_tags = len(tags)

    and even a simple::

       #str() iterates over all tags to construct a space separated list
       print(str(tags))

    will "exhaust" the Tags. If you need to re-iterate over a list of
    tags you will need to retrieve a new :class:`Tags` object.
    """

    #notmuch_tags_get
    _get = nmlib.notmuch_tags_get
    _get.argtypes = [NotmuchTagsP]
    _get.restype = c_char_p

    def __init__(self, tags_p, parent=None):
        """
        :param tags_p: A pointer to an underlying *notmuch_tags_t*
             structure. These are not publicly exposed, so a user
             will almost never instantiate a :class:`Tags` object
             herself. They are usually handed back as a result,
             e.g. in :meth:`Database.get_all_tags`.  *tags_p* must be
             valid, we will raise an :exc:`NullPointerError` if it is
             `None`.
        :type tags_p: :class:`ctypes.c_void_p`
        :param parent: The parent object (ie :class:`Database` or
             :class:`Message` these tags are derived from, and saves a
             reference to it, so we can automatically delete the db object
             once all derived objects are dead.
        :TODO: Make the iterator optionally work more than once by
               cache the tags in the Python object(?)
        """
        if not tags_p:
            raise NullPointerError()

        self._tags = tags_p
        #save reference to parent object so we keep it alive
        self._parent = parent

    def __iter__(self):
        """ Make Tags an iterator """
        return self

    _valid = nmlib.notmuch_tags_valid
    _valid.argtypes = [NotmuchTagsP]
    _valid.restype = bool

    _move_to_next = nmlib.notmuch_tags_move_to_next
    _move_to_next.argtypes = [NotmuchTagsP]
    _move_to_next.restype = None

    def __next__(self):
        if not self._tags:
            raise NotInitializedError()
        if not self._valid(self._tags):
            self._tags = None
            raise StopIteration
        tag = Tags._get(self._tags).decode('UTF-8')
        self._move_to_next(self._tags)
        return tag
    next = __next__ # python2.x iterator protocol compatibility

    def __nonzero__(self):
        '''
        Implement truth value testing. If __nonzero__ is not
        implemented, the python runtime would fall back to `len(..) >
        0` thus exhausting the iterator.

        :returns: True if the wrapped iterator has at least one more object
                  left.
        '''
        return self._tags and self._valid(self._tags)

    def __unicode__(self):
        """string representation of :class:`Tags`: a space separated list of tags

        .. note::

            As this iterates over the tags, we will not be able to iterate over
            them again (as in retrieve them)! If the tags have been exhausted
            already, this will raise a :exc:`NotInitializedError`on subsequent
            attempts.
        """
        return " ".join(self)

    _destroy = nmlib.notmuch_tags_destroy
    _destroy.argtypes = [NotmuchTagsP]
    _destroy.restype = None

    def __del__(self):
        """Close and free the notmuch tags"""
        if self._tags:
            self._destroy(self._tags)



import email.message
import mailbox
import pathlib
import shutil
import socket
import subprocess
import textwrap
import time
import os

import pytest


def pytest_report_header():
    which = shutil.which('notmuch')
    vers = subprocess.run(['notmuch', '--version'], stdout=subprocess.PIPE)
    return ['{} ({})'.format(vers.stdout.decode(errors='replace').strip(),which)]


@pytest.fixture(scope='function')
def tmppath(tmpdir):
    """The tmpdir fixture wrapped in pathlib.Path."""
    return pathlib.Path(str(tmpdir))


@pytest.fixture
def notmuch(maildir):
    """Return a function which runs notmuch commands on our test maildir.

    This uses the notmuch-config file created by the ``maildir``
    fixture.
    """
    def run(*args):
        """Run a notmuch command.

        This function runs with a timeout error as many notmuch
        commands may block if multiple processes are trying to open
        the database in write-mode.  It is all too easy to
        accidentally do this in the unittests.
        """
        cfg_fname = maildir.path / 'notmuch-config'
        cmd = ['notmuch'] + list(args)
        env = os.environ.copy()
        env['NOTMUCH_CONFIG'] = str(cfg_fname)
        proc = subprocess.run(cmd,
                              timeout=120,
                              env=env)
        proc.check_returncode()
    return run


@pytest.fixture
def maildir(tmppath):
    """A basic test interface to a valid maildir directory.

    This creates a valid maildir and provides a simple mechanism to
    deliver test emails to it.  It also writes a notmuch-config file
    in the top of the maildir.
    """
    cur = tmppath / 'cur'
    cur.mkdir()
    new = tmppath / 'new'
    new.mkdir()
    tmp = tmppath / 'tmp'
    tmp.mkdir()
    cfg_fname = tmppath/'notmuch-config'
    with cfg_fname.open('w') as fp:
        fp.write(textwrap.dedent("""\
            [database]
            path={tmppath!s}
            [user]
            name=Some Hacker
            primary_email=dst@example.com
            [new]
            tags=unread;inbox;
            ignore=
            [search]
            exclude_tags=deleted;spam;
            [maildir]
            synchronize_flags=true
            """.format(tmppath=tmppath)))
    return MailDir(tmppath)


class MailDir:
    """An interface around a correct maildir."""

    def __init__(self, path):
        self._path = pathlib.Path(path)
        self.mailbox = mailbox.Maildir(str(path))
        self._idcount = 0

    @property
    def path(self):
        """The pathname of the maildir."""
        return self._path

    def _next_msgid(self):
        """Return a new unique message ID."""
        msgid = '{}@{}'.format(self._idcount, socket.getfqdn())
        self._idcount += 1
        return msgid

    def deliver(self,
                subject='Test mail',
                body='This is a test mail',
                to='dst@example.com',
                frm='src@example.com',
                headers=None,
                new=False,      # Move to new dir or cur dir?
                keywords=None,  # List of keywords or labels
                seen=False,     # Seen flag (cur dir only)
                replied=False,  # Replied flag (cur dir only)
                flagged=False):  # Flagged flag (cur dir only)
        """Deliver a new mail message in the mbox.

        This does only adds the message to maildir, does not insert it
        into the notmuch database.

        :returns: A tuple of (msgid, pathname).
        """
        msgid = self._next_msgid()
        when = time.time()
        msg = email.message.EmailMessage()
        msg.add_header('Received', 'by MailDir; {}'.format(time.ctime(when)))
        msg.add_header('Message-ID', '<{}>'.format(msgid))
        msg.add_header('Date', time.ctime(when))
        msg.add_header('From', frm)
        msg.add_header('To', to)
        msg.add_header('Subject', subject)
        if headers:
            for h, v in headers:
                msg.add_header(h, v)
        msg.set_content(body)
        mdmsg = mailbox.MaildirMessage(msg)
        if not new:
            mdmsg.set_subdir('cur')
        if flagged:
            mdmsg.add_flag('F')
        if replied:
            mdmsg.add_flag('R')
        if seen:
            mdmsg.add_flag('S')
        boxid = self.mailbox.add(mdmsg)
        basename = boxid
        if mdmsg.get_info():
            basename += mailbox.Maildir.colon + mdmsg.get_info()
        msgpath = self.path / mdmsg.get_subdir() / basename
        return (msgid, msgpath)



import collections.abc
import time

import pytest

import notmuch2


@pytest.fixture
def thread(maildir, notmuch):
    """Return a single thread with one matched message."""
    msgid, _ = maildir.deliver(body='foo')
    maildir.deliver(body='bar',
                    headers=[('In-Reply-To', '<{}>'.format(msgid))])
    notmuch('new')
    with notmuch2.Database(maildir.path, config=notmuch2.Database.CONFIG.EMPTY) as db:
        yield next(db.threads('foo'))


def test_type(thread):
    assert isinstance(thread, notmuch2.Thread)
    assert isinstance(thread, collections.abc.Iterable)


def test_threadid(thread):
    assert isinstance(thread.threadid, notmuch2.BinString)
    assert thread.threadid


def test_len(thread):
    assert len(thread) == 2


def test_toplevel_type(thread):
    assert isinstance(thread.toplevel(), collections.abc.Iterator)


def test_toplevel(thread):
    msgs = thread.toplevel()
    assert isinstance(next(msgs), notmuch2.Message)
    with pytest.raises(StopIteration):
        next(msgs)


def test_toplevel_reply(thread):
    msg = next(thread.toplevel())
    assert isinstance(next(msg.replies()), notmuch2.Message)


def test_iter(thread):
    msgs = list(iter(thread))
    assert len(msgs) == len(thread)
    for msg in msgs:
        assert isinstance(msg, notmuch2.Message)


def test_matched(thread):
    assert thread.matched == 1

def test_matched_iter(thread):
    count = 0
    msgs = list(iter(thread))
    for msg in msgs:
        if msg.matched:
            count += 1
    assert count == thread.matched

def test_authors_type(thread):
    assert isinstance(thread.authors, notmuch2.BinString)


def test_authors(thread):
    assert thread.authors == 'src@example.com'


def test_subject(thread):
    assert thread.subject == 'Test mail'


def test_first(thread):
    # XXX Someone seems to treat things as local time instead of
    #     UTC or the other way around.
    now = int(time.time())
    assert abs(now - thread.first) < 3600*24


def test_last(thread):
    # XXX Someone seems to treat things as local time instead of
    #     UTC or the other way around.
    now = int(time.time())
    assert abs(now - thread.last) < 3600*24


def test_first_last(thread):
    # Sadly we only have second resolution so these will always be the
    # same time in our tests.
    assert thread.first <= thread.last


def test_tags_type(thread):
    assert isinstance(thread.tags, notmuch2.ImmutableTagSet)


def test_tags_cache(thread):
    assert thread.tags is thread.tags


def test_tags(thread):
    assert 'inbox' in thread.tags



"""
This file is part of notmuch.

Notmuch is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

Notmuch is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with notmuch.  If not, see <https://www.gnu.org/licenses/>.

Copyright 2010 Sebastian Spaeth <Sebastian@SSpaeth.de>
"""
from ctypes import c_char_p
from .globals import (
    nmlib,
    NotmuchFilenamesP,
    Python3StringMixIn,
)
from .errors import (
    NullPointerError,
    NotInitializedError,
)


class Filenames(Python3StringMixIn):
    """Represents a list of filenames as returned by notmuch

    Objects of this class implement the iterator protocol.

    .. note::

        The underlying library only provides a one-time iterator (it
        cannot reset the iterator to the start). Thus iterating over
        the function will "exhaust" the list of tags, and a subsequent
        iteration attempt will raise a
        :exc:`NotInitializedError`. Also note, that any function that
        uses iteration (nearly all) will also exhaust the tags. So
        both::

           for name in filenames: print name

        as well as::

           list_of_names = list(names)

        and even a simple::

           #str() iterates over all tags to construct a space separated list
           print(str(filenames))

        will "exhaust" the Filenames. However, you can use
        :meth:`Message.get_filenames` repeatedly to get fresh
        Filenames objects to perform various actions on filenames.
    """

    #notmuch_filenames_get
    _get = nmlib.notmuch_filenames_get
    _get.argtypes = [NotmuchFilenamesP]
    _get.restype = c_char_p

    def __init__(self, files_p, parent):
        """
        :param files_p: A pointer to an underlying *notmuch_tags_t*
             structure. These are not publicly exposed, so a user
             will almost never instantiate a :class:`Tags` object
             herself. They are usually handed back as a result,
             e.g. in :meth:`Database.get_all_tags`.  *tags_p* must be
             valid, we will raise an :exc:`NullPointerError`
             if it is `None`.
        :type files_p: :class:`ctypes.c_void_p`
        :param parent: The parent object (ie :class:`Message` these
             filenames are derived from, and saves a
             reference to it, so we can automatically delete the db object
             once all derived objects are dead.
        """
        if not files_p:
            raise NullPointerError()

        self._files_p = files_p
        #save reference to parent object so we keep it alive
        self._parent = parent

    def __iter__(self):
        """ Make Filenames an iterator """
        return self

    _valid = nmlib.notmuch_filenames_valid
    _valid.argtypes = [NotmuchFilenamesP]
    _valid.restype = bool

    _move_to_next = nmlib.notmuch_filenames_move_to_next
    _move_to_next.argtypes = [NotmuchFilenamesP]
    _move_to_next.restype = None

    def __next__(self):
        if not self._files_p:
            raise NotInitializedError()

        if not self._valid(self._files_p):
            self._files_p = None
            raise StopIteration

        file_ = Filenames._get(self._files_p)
        self._move_to_next(self._files_p)
        return file_.decode('utf-8', 'ignore')
    next = __next__ # python2.x iterator protocol compatibility

    def __unicode__(self):
        """Represent Filenames() as newline-separated list of full paths

        .. note::

            This method exhausts the iterator object, so you will not be able to
            iterate over them again.
        """
        return "\n".join(self)

    _destroy = nmlib.notmuch_filenames_destroy
    _destroy.argtypes = [NotmuchFilenamesP]
    _destroy.restype = None

    def __del__(self):
        """Close and free the notmuch filenames"""
        if self._files_p:
            self._destroy(self._files_p)



import collections.abc

import notmuch2._base as base
import notmuch2._capi as capi
import notmuch2._errors as errors


__all__ = ['ConfigMapping']


class ConfigIter(base.NotmuchIter):

    def __init__(self, parent, iter_p):
        super().__init__(
            parent, iter_p,
            fn_destroy=capi.lib.notmuch_config_pairs_destroy,
            fn_valid=capi.lib.notmuch_config_pairs_valid,
            fn_get=capi.lib.notmuch_config_pairs_key,
            fn_next=capi.lib.notmuch_config_pairs_move_to_next)

    def __next__(self):
        # skip pairs whose value is NULL
        while capi.lib.notmuch_config_pairs_valid(super()._iter_p):
            val_p = capi.lib.notmuch_config_pairs_value(super()._iter_p)
            key_p = capi.lib.notmuch_config_pairs_key(super()._iter_p)
            if key_p == capi.ffi.NULL:
                # this should never happen
                raise errors.NullPointerError
            key = base.BinString.from_cffi(key_p)
            capi.lib.notmuch_config_pairs_move_to_next(super()._iter_p)
            if val_p != capi.ffi.NULL and base.BinString.from_cffi(val_p) != "":
                return key
        self._destroy()
        raise StopIteration

class ConfigMapping(base.NotmuchObject, collections.abc.MutableMapping):
    """The config key/value pairs loaded from the database, config file,
    and and/or defaults.

    The entries are exposed as a :class:`collections.abc.MutableMapping` object.
    Note that setting a value to an empty string is the same as deleting it.

    Mutating (deleting or updating values) in the map persists only in
    the database, which can be shadowed by config files.

    :param parent: the parent object
    :param ptr_name: the name of the attribute on the parent which will
       return the memory pointer.  This allows this object to
       access the pointer via the parent's descriptor and thus
       trigger :class:`MemoryPointer`'s memory safety.

    """

    def __init__(self, parent, ptr_name):
        self._parent = parent
        self._ptr = lambda: getattr(parent, ptr_name)

    @property
    def alive(self):
        return self._parent.alive

    def _destroy(self):
        pass

    def __getitem__(self, key):
        if isinstance(key, str):
            key = key.encode('utf-8')
        val_pp = capi.ffi.new('char**')
        ret = capi.lib.notmuch_database_get_config(self._ptr(), key, val_pp)
        if ret != capi.lib.NOTMUCH_STATUS_SUCCESS:
            raise errors.NotmuchError(ret)
        val = base.BinString.from_cffi(val_pp[0])
        capi.lib.free(val_pp[0])
        if val == '':
            raise KeyError
        return val

    def __setitem__(self, key, val):
        if isinstance(key, str):
            key = key.encode('utf-8')
        if isinstance(val, str):
            val = val.encode('utf-8')
        ret = capi.lib.notmuch_database_set_config(self._ptr(), key, val)
        if ret != capi.lib.NOTMUCH_STATUS_SUCCESS:
            raise errors.NotmuchError(ret)

    def __delitem__(self, key):
        self[key] = ""

    def __iter__(self):
        """Return an iterator over the config items.

        :raises NullPointerError: If the iterator can not be created.
        """
        config_pairs_p = capi.lib.notmuch_config_get_pairs(self._ptr(), b'')
        if config_pairs_p == capi.ffi.NULL:
            raise KeyError
        return ConfigIter(self._parent, config_pairs_p)

    def __len__(self):
        return sum(1 for t in self)



from notmuch2 import _capi as capi
from notmuch2 import _errors as errors

def test_status_no_message():
    exc = errors.NotmuchError(capi.lib.NOTMUCH_STATUS_PATH_ERROR)
    assert exc.status == capi.lib.NOTMUCH_STATUS_PATH_ERROR
    assert exc.message is None
    assert str(exc) == 'Path supplied is illegal for this function'



import collections.abc
import time
import pathlib

import pytest

import notmuch2


class TestMessage:
    MaildirMsg = collections.namedtuple('MaildirMsg', ['msgid', 'path'])

    @pytest.fixture
    def maildir_msg(self, maildir):
        msgid, path = maildir.deliver()
        return self.MaildirMsg(msgid, path)

    @pytest.fixture
    def db(self, maildir):
        with notmuch2.Database.create(maildir.path) as db:
            yield db

    @pytest.fixture
    def msg(self, db, maildir_msg):
        msg, dup = db.add(maildir_msg.path, sync_flags=False)
        yield msg

    def test_type(self, msg):
        assert isinstance(msg, notmuch2.NotmuchObject)
        assert isinstance(msg, notmuch2.Message)

    def test_alive(self, msg):
        assert msg.alive

    def test_hash(self, msg):
        assert hash(msg)

    def test_eq(self, db, msg):
        copy = db.get(msg.path)
        assert msg == copy

    def test_messageid_type(self, msg):
        assert isinstance(msg.messageid, str)
        assert isinstance(msg.messageid, notmuch2.BinString)
        assert isinstance(bytes(msg.messageid), bytes)

    def test_messageid(self, msg, maildir_msg):
        assert msg.messageid == maildir_msg.msgid

    def test_messageid_find(self, db, msg):
        copy = db.find(msg.messageid)
        assert msg.messageid == copy.messageid

    def test_threadid_type(self, msg):
        assert isinstance(msg.threadid, str)
        assert isinstance(msg.threadid, notmuch2.BinString)
        assert isinstance(bytes(msg.threadid), bytes)

    def test_path_type(self, msg):
        assert isinstance(msg.path, pathlib.Path)

    def test_path(self, msg, maildir_msg):
        assert msg.path == maildir_msg.path

    def test_pathb_type(self, msg):
        assert isinstance(msg.pathb, bytes)

    def test_pathb(self, msg, maildir_msg):
        assert msg.path == maildir_msg.path

    def test_filenames_type(self, msg):
        ifn = msg.filenames()
        assert isinstance(ifn, collections.abc.Iterator)

    def test_filenames(self, msg):
        ifn = msg.filenames()
        fn = next(ifn)
        assert fn == msg.path
        assert isinstance(fn, pathlib.Path)
        with pytest.raises(StopIteration):
            next(ifn)
        assert list(msg.filenames()) == [msg.path]

    def test_filenamesb_type(self, msg):
        ifn = msg.filenamesb()
        assert isinstance(ifn, collections.abc.Iterator)

    def test_filenamesb(self, msg):
        ifn = msg.filenamesb()
        fn = next(ifn)
        assert fn == msg.pathb
        assert isinstance(fn, bytes)
        with pytest.raises(StopIteration):
            next(ifn)
        assert list(msg.filenamesb()) == [msg.pathb]

    def test_ghost_no(self, msg):
        assert not msg.ghost

    def test_matched_no(self,msg):
        assert not msg.matched

    def test_date(self, msg):
        # XXX Someone seems to treat things as local time instead of
        #     UTC or the other way around.
        now = int(time.time())
        assert abs(now - msg.date) < 3600*24

    def test_header(self, msg):
        assert msg.header('from') == 'src@example.com'

    def test_header_not_present(self, msg):
        with pytest.raises(LookupError):
            msg.header('foo')

    def test_freeze(self, msg):
        with msg.frozen():
            msg.tags.add('foo')
            msg.tags.add('bar')
            msg.tags.discard('foo')
        assert 'foo' not in msg.tags
        assert 'bar' in msg.tags

    def test_freeze_err(self, msg):
        msg.tags.add('foo')
        try:
            with msg.frozen():
                msg.tags.clear()
                raise Exception('oops')
        except Exception:
            assert 'foo' in msg.tags
        else:
            pytest.fail('Context manager did not raise')

    def test_replies_type(self, msg):
        assert isinstance(msg.replies(), collections.abc.Iterator)

    def test_replies(self, msg):
        with pytest.raises(StopIteration):
            next(msg.replies())


class TestProperties:

    @pytest.fixture
    def props(self, maildir):
        msgid, path = maildir.deliver()
        with notmuch2.Database.create(maildir.path) as db:
            msg, dup = db.add(path, sync_flags=False)
            yield msg.properties

    def test_type(self, props):
        assert isinstance(props, collections.abc.MutableMapping)

    def test_add_single(self, props):
        props['foo'] = 'bar'
        assert props['foo'] == 'bar'
        props.add('bar', 'baz')
        assert props['bar'] == 'baz'

    def test_add_dup(self, props):
        props.add('foo', 'bar')
        props.add('foo', 'baz')
        assert props['foo'] == 'bar'
        assert (set(props.getall('foo', exact=True))
                == {('foo', 'bar'), ('foo', 'baz')})

    def test_len(self, props):
        props.add('foo', 'a')
        props.add('foo', 'b')
        props.add('bar', 'a')
        assert len(props) == 3
        assert len(props.keys()) == 2
        assert len(props.values()) == 2
        assert len(props.items()) == 3

    def test_del(self, props):
        props.add('foo', 'a')
        props.add('foo', 'b')
        del props['foo']
        with pytest.raises(KeyError):
            props['foo']

    def test_remove(self, props):
        props.add('foo', 'a')
        props.add('foo', 'b')
        props.remove('foo', 'a')
        assert props['foo'] == 'b'

    def test_view_abcs(self, props):
        assert isinstance(props.keys(), collections.abc.KeysView)
        assert isinstance(props.values(), collections.abc.ValuesView)
        assert isinstance(props.items(), collections.abc.ItemsView)

    def test_pop(self, props):
        props.add('foo', 'a')
        props.add('foo', 'b')
        val = props.pop('foo')
        assert val == 'a'

    def test_pop_default(self, props):
        with pytest.raises(KeyError):
            props.pop('foo')
        assert props.pop('foo', 'default') == 'default'

    def test_popitem(self, props):
        props.add('foo', 'a')
        assert props.popitem() == ('foo', 'a')
        with pytest.raises(KeyError):
            props.popitem()

    def test_clear(self, props):
        props.add('foo', 'a')
        props.clear()
        assert len(props) == 0

    def test_getall(self, props):
        props.add('foo', 'a')
        assert set(props.getall('foo')) == {('foo', 'a')}

    def test_getall_prefix(self, props):
        props.add('foo', 'a')
        props.add('foobar', 'b')
        assert set(props.getall('foo')) == {('foo', 'a'), ('foobar', 'b')}

    def test_getall_exact(self, props):
        props.add('foo', 'a')
        props.add('foobar', 'b')
        assert set(props.getall('foo', exact=True)) == {('foo', 'a')}



import collections.abc

import pytest

import notmuch2._database as dbmod

import notmuch2._config as config


class TestIter:

    @pytest.fixture
    def db(self, maildir):
        with dbmod.Database.create(maildir.path) as db:
            yield db

    def test_type(self, db):
        assert isinstance(db.config, collections.abc.MutableMapping)
        assert isinstance(db.config, config.ConfigMapping)

    def test_alive(self, db):
        assert db.config.alive

    def test_set_get(self, maildir):
        # Ensure get-set works from different db objects
        with dbmod.Database.create(maildir.path, config=dbmod.Database.CONFIG.EMPTY) as db0:
            db0.config['spam'] = 'ham'
        with dbmod.Database(maildir.path, config=dbmod.Database.CONFIG.EMPTY) as db1:
            assert db1.config['spam'] == 'ham'

    def test_get_keyerror(self, db):
        with pytest.raises(KeyError):
            val = db.config['not-a-key']
            print(repr(val))

    def test_iter(self, db):
        def has_prefix(x):
            return x.startswith('TEST.')

        assert [ x for x in db.config if has_prefix(x) ] == []
        db.config['TEST.spam'] = 'TEST.ham'
        db.config['TEST.eggs'] = 'TEST.bacon'
        assert { x for x in db.config if has_prefix(x) } == {'TEST.spam', 'TEST.eggs'}
        assert { x for x in db.config.keys() if has_prefix(x) } == {'TEST.spam', 'TEST.eggs'}
        assert { x for x in db.config.values() if has_prefix(x) } == {'TEST.ham', 'TEST.bacon'}
        assert { (x, y) for (x,y) in db.config.items() if has_prefix(x) } == \
            {('TEST.spam', 'TEST.ham'), ('TEST.eggs', 'TEST.bacon')}

    def test_len(self, db):
        defaults = len(db.config)
        db.config['spam'] = 'ham'
        assert len(db.config) == defaults + 1
        db.config['eggs'] = 'bacon'
        assert len(db.config) == defaults + 2

    def test_del(self, db):
        db.config['spam'] = 'ham'
        assert db.config.get('spam') == 'ham'
        del db.config['spam']
        assert db.config.get('spam') is None



#!/usr/bin/env python3
import re
import sys
import json


EXPR_RE = re.compile('(?P<label>[a-zA-Z0-9_-]+):(?P<address>[^=!]+)(?:(?P<type>[=!])(?P<val>.*))?', re.DOTALL|re.MULTILINE)


if len(sys.argv) < 2:
    sys.exit('usage: '+ sys.argv[0] + """ EXPR [EXPR]

Takes json data on stdin and evaluates test expressions specified in
arguments.  Each test is evaluated, and output is printed only if the
test fails.  If any test fails the return value of execution will be
non-zero.

EXPR can be one of following types:

Value test: test that object in json data found at address is equal to
specified value:

  label:address=value

Existence test: test that dict or list in json data found at address
does *not* contain the specified key:

  label:address!key

Extract: extract object from json data found at address and print

  label:address

Results are printed to stdout prefixed by expression label.  In all
cases the test will fail if object does not exist in data.

Example:

0 $ echo '["a", "b", {"c": 1}]' | python3 json_check_nodes.py 'second_d:[1]="d"' 'no_c:[2]!"c"'
second_d: value not equal: data[1] = 'b' != 'd'
no_c: dict contains key: data[2]["c"] = 1
1 $

""")


# parse expressions from arguments
exprs = []
for expr in sys.argv[1:]:
    m = re.match(EXPR_RE, expr)
    if not m:
        sys.exit("Invalid expression: {}".format(expr))
    exprs.append(m)

data = json.load(sys.stdin)

fail = False

for expr in exprs:
    # print(expr.groups(),fail)

    e = 'data{}'.format(expr.group('address'))
    try:
        val = eval(e)
    except SyntaxError:
        fail = True
        print("{}: syntax error on evaluation of object: {}".format(
            expr.group('label'), e))
        continue
    except:
        fail = True
        print("{}: object not found: data{}".format(
            expr.group('label'), expr.group('address')))
        continue

    if expr.group('type') == '=':
        try:
            obj_val = json.loads(expr.group('val'))
        except:
            fail = True
            print("{}: error evaluating value: {}".format(
                expr.group('label'), expr.group('address')))
            continue
        if val != obj_val:
            fail = True
            print("{}: value not equal: data{} = {} != {}".format(
                expr.group('label'), expr.group('address'), repr(val), repr(obj_val)))

    elif expr.group('type') == '!':
        if not isinstance(val, (dict, list)):
            fail = True
            print("{}: not a dict or a list: data{}".format(
                expr.group('label'), expr.group('address')))
            continue
        try:
            idx = json.loads(expr.group('val'))
            if idx in val:
                fail = True
                print("{}: {} contains key: {}[{}] = {}".format(
                    expr.group('label'), type(val).__name__, e, expr.group('val'), val[idx]))
        except SyntaxError:
            fail = True
            print("{}: syntax error on evaluation of value: {}".format(
                expr.group('label'), expr.group('val')))
            continue


    elif expr.group('type') is None:
        print("{}: {}".format(expr.group('label'), val))


if fail:
    sys.exit(1)
sys.exit(0)



import cffi
from _notmuch_config import *

ffibuilder = cffi.FFI()
ffibuilder.set_source(
    'notmuch2._capi',
    r"""
    #include <stdlib.h>
    #include <time.h>
    #include <notmuch.h>

    #if LIBNOTMUCH_MAJOR_VERSION < 5
        #error libnotmuch version not supported by notmuch2 python bindings
    #endif
    #if LIBNOTMUCH_MINOR_VERSION < 1
        #ERROR libnotmuch  version < 5.1 not supported
    #endif
    """,
    include_dirs=[NOTMUCH_INCLUDE_DIR],
    library_dirs=[NOTMUCH_LIB_DIR],
    libraries=['notmuch'],
)
ffibuilder.cdef(
    r"""
    void free(void *ptr);
    typedef int... time_t;

    #define LIBNOTMUCH_MAJOR_VERSION ...
    #define LIBNOTMUCH_MINOR_VERSION ...
    #define LIBNOTMUCH_MICRO_VERSION ...

    #define NOTMUCH_TAG_MAX ...

    typedef enum _notmuch_status {
        NOTMUCH_STATUS_SUCCESS = 0,
        NOTMUCH_STATUS_OUT_OF_MEMORY,
        NOTMUCH_STATUS_READ_ONLY_DATABASE,
        NOTMUCH_STATUS_XAPIAN_EXCEPTION,
        NOTMUCH_STATUS_FILE_ERROR,
        NOTMUCH_STATUS_FILE_NOT_EMAIL,
        NOTMUCH_STATUS_DUPLICATE_MESSAGE_ID,
        NOTMUCH_STATUS_NULL_POINTER,
        NOTMUCH_STATUS_TAG_TOO_LONG,
        NOTMUCH_STATUS_UNBALANCED_FREEZE_THAW,
        NOTMUCH_STATUS_UNBALANCED_ATOMIC,
        NOTMUCH_STATUS_UNSUPPORTED_OPERATION,
        NOTMUCH_STATUS_UPGRADE_REQUIRED,
        NOTMUCH_STATUS_PATH_ERROR,
        NOTMUCH_STATUS_ILLEGAL_ARGUMENT,
        NOTMUCH_STATUS_MALFORMED_CRYPTO_PROTOCOL,
        NOTMUCH_STATUS_FAILED_CRYPTO_CONTEXT_CREATION,
        NOTMUCH_STATUS_UNKNOWN_CRYPTO_PROTOCOL,
        NOTMUCH_STATUS_NO_CONFIG,
        NOTMUCH_STATUS_NO_DATABASE,
        NOTMUCH_STATUS_DATABASE_EXISTS,
        NOTMUCH_STATUS_BAD_QUERY_SYNTAX,
        NOTMUCH_STATUS_NO_MAIL_ROOT,
        NOTMUCH_STATUS_CLOSED_DATABASE,
        NOTMUCH_STATUS_LAST_STATUS
    } notmuch_status_t;
    typedef enum {
        NOTMUCH_DATABASE_MODE_READ_ONLY = 0,
        NOTMUCH_DATABASE_MODE_READ_WRITE
    } notmuch_database_mode_t;
    typedef int notmuch_bool_t;
    typedef enum _notmuch_message_flag {
        NOTMUCH_MESSAGE_FLAG_MATCH,
        NOTMUCH_MESSAGE_FLAG_EXCLUDED,
        NOTMUCH_MESSAGE_FLAG_GHOST,
    } notmuch_message_flag_t;
    typedef enum {
        NOTMUCH_SORT_OLDEST_FIRST,
        NOTMUCH_SORT_NEWEST_FIRST,
        NOTMUCH_SORT_MESSAGE_ID,
        NOTMUCH_SORT_UNSORTED
    } notmuch_sort_t;
    typedef enum {
        NOTMUCH_EXCLUDE_FLAG,
        NOTMUCH_EXCLUDE_TRUE,
        NOTMUCH_EXCLUDE_FALSE,
        NOTMUCH_EXCLUDE_ALL
    } notmuch_exclude_t;
    typedef enum {
        NOTMUCH_DECRYPT_FALSE,
        NOTMUCH_DECRYPT_TRUE,
        NOTMUCH_DECRYPT_AUTO,
        NOTMUCH_DECRYPT_NOSTASH,
    } notmuch_decryption_policy_t;

    // These are fully opaque types for us, we only ever use pointers.
    typedef struct _notmuch_database notmuch_database_t;
    typedef struct _notmuch_query notmuch_query_t;
    typedef struct _notmuch_threads notmuch_threads_t;
    typedef struct _notmuch_thread notmuch_thread_t;
    typedef struct _notmuch_messages notmuch_messages_t;
    typedef struct _notmuch_message notmuch_message_t;
    typedef struct _notmuch_tags notmuch_tags_t;
    typedef struct _notmuch_string_map_iterator notmuch_message_properties_t;
    typedef struct _notmuch_directory notmuch_directory_t;
    typedef struct _notmuch_filenames notmuch_filenames_t;
    typedef struct _notmuch_config_pairs notmuch_config_pairs_t;
    typedef struct _notmuch_indexopts notmuch_indexopts_t;

    const char *
    notmuch_status_to_string (notmuch_status_t status);

    notmuch_status_t
    notmuch_database_create_with_config (const char *database_path,
                                         const char *config_path,
                                         const char *profile,
                                         notmuch_database_t **database,
                                         char **error_message);
    notmuch_status_t
    notmuch_database_open_with_config (const char *database_path,
                                       notmuch_database_mode_t mode,
                                       const char *config_path,
                                       const char *profile,
                                       notmuch_database_t **database,
                                       char **error_message);
    notmuch_status_t
    notmuch_database_close (notmuch_database_t *database);
    notmuch_status_t
    notmuch_database_destroy (notmuch_database_t *database);
    const char *
    notmuch_database_get_path (notmuch_database_t *database);
    unsigned int
    notmuch_database_get_version (notmuch_database_t *database);
    notmuch_bool_t
    notmuch_database_needs_upgrade (notmuch_database_t *database);
    notmuch_status_t
    notmuch_database_begin_atomic (notmuch_database_t *notmuch);
    notmuch_status_t
    notmuch_database_end_atomic (notmuch_database_t *notmuch);
    unsigned long
    notmuch_database_get_revision (notmuch_database_t *notmuch,
                                   const char **uuid);
    notmuch_status_t
    notmuch_database_index_file (notmuch_database_t *database,
                                 const char *filename,
                                 notmuch_indexopts_t *indexopts,
                                 notmuch_message_t **message);
    notmuch_status_t
    notmuch_database_remove_message (notmuch_database_t *database,
                                     const char *filename);
    notmuch_status_t
    notmuch_database_find_message (notmuch_database_t *database,
                                   const char *message_id,
                                   notmuch_message_t **message);
    notmuch_status_t
    notmuch_database_find_message_by_filename (notmuch_database_t *notmuch,
                                               const char *filename,
                                               notmuch_message_t **message);
    notmuch_tags_t *
    notmuch_database_get_all_tags (notmuch_database_t *db);

    notmuch_query_t *
    notmuch_query_create (notmuch_database_t *database,
                          const char *query_string);
    const char *
    notmuch_query_get_query_string (const notmuch_query_t *query);
    notmuch_database_t *
    notmuch_query_get_database (const notmuch_query_t *query);
    void
    notmuch_query_set_omit_excluded (notmuch_query_t *query,
                                     notmuch_exclude_t omit_excluded);
    void
    notmuch_query_set_sort (notmuch_query_t *query, notmuch_sort_t sort);
    notmuch_sort_t
    notmuch_query_get_sort (const notmuch_query_t *query);
    notmuch_status_t
    notmuch_query_add_tag_exclude (notmuch_query_t *query, const char *tag);
    notmuch_status_t
    notmuch_query_search_threads (notmuch_query_t *query,
                                  notmuch_threads_t **out);
    notmuch_status_t
    notmuch_query_search_messages (notmuch_query_t *query,
                                   notmuch_messages_t **out);
    notmuch_status_t
    notmuch_query_count_messages (notmuch_query_t *query, unsigned int *count);
    notmuch_status_t
    notmuch_query_count_threads (notmuch_query_t *query, unsigned *count);
    void
    notmuch_query_destroy (notmuch_query_t *query);

    notmuch_bool_t
    notmuch_threads_valid (notmuch_threads_t *threads);
    notmuch_thread_t *
    notmuch_threads_get (notmuch_threads_t *threads);
    void
    notmuch_threads_move_to_next (notmuch_threads_t *threads);
    void
    notmuch_threads_destroy (notmuch_threads_t *threads);

    const char *
    notmuch_thread_get_thread_id (notmuch_thread_t *thread);
    notmuch_messages_t *
    notmuch_message_get_replies (notmuch_message_t *message);
    int
    notmuch_thread_get_total_messages (notmuch_thread_t *thread);
    notmuch_messages_t *
    notmuch_thread_get_toplevel_messages (notmuch_thread_t *thread);
    notmuch_messages_t *
    notmuch_thread_get_messages (notmuch_thread_t *thread);
    int
    notmuch_thread_get_matched_messages (notmuch_thread_t *thread);
    const char *
    notmuch_thread_get_authors (notmuch_thread_t *thread);
    const char *
    notmuch_thread_get_subject (notmuch_thread_t *thread);
    time_t
    notmuch_thread_get_oldest_date (notmuch_thread_t *thread);
    time_t
    notmuch_thread_get_newest_date (notmuch_thread_t *thread);
    notmuch_tags_t *
    notmuch_thread_get_tags (notmuch_thread_t *thread);
    void
    notmuch_thread_destroy (notmuch_thread_t *thread);

    notmuch_bool_t
    notmuch_messages_valid (notmuch_messages_t *messages);
    notmuch_message_t *
    notmuch_messages_get (notmuch_messages_t *messages);
    void
    notmuch_messages_move_to_next (notmuch_messages_t *messages);
    void
    notmuch_messages_destroy (notmuch_messages_t *messages);
    notmuch_tags_t *
    notmuch_messages_collect_tags (notmuch_messages_t *messages);

    const char *
    notmuch_message_get_message_id (notmuch_message_t *message);
    const char *
    notmuch_message_get_thread_id (notmuch_message_t *message);
    const char *
    notmuch_message_get_filename (notmuch_message_t *message);
    notmuch_filenames_t *
    notmuch_message_get_filenames (notmuch_message_t *message);
    notmuch_bool_t
    notmuch_message_get_flag (notmuch_message_t *message,
                              notmuch_message_flag_t flag);
    void
    notmuch_message_set_flag (notmuch_message_t *message,
                              notmuch_message_flag_t flag,
                              notmuch_bool_t value);
    time_t
    notmuch_message_get_date  (notmuch_message_t *message);
    const char *
    notmuch_message_get_header (notmuch_message_t *message,
                                const char *header);
    notmuch_tags_t *
    notmuch_message_get_tags (notmuch_message_t *message);
    notmuch_status_t
    notmuch_message_add_tag (notmuch_message_t *message, const char *tag);
    notmuch_status_t
    notmuch_message_remove_tag (notmuch_message_t *message, const char *tag);
    notmuch_status_t
    notmuch_message_remove_all_tags (notmuch_message_t *message);
    notmuch_status_t
    notmuch_message_maildir_flags_to_tags (notmuch_message_t *message);
    notmuch_status_t
    notmuch_message_tags_to_maildir_flags (notmuch_message_t *message);
    notmuch_status_t
    notmuch_message_freeze (notmuch_message_t *message);
    notmuch_status_t
    notmuch_message_thaw (notmuch_message_t *message);
    notmuch_status_t
    notmuch_message_get_property (notmuch_message_t *message,
                                  const char *key, const char **value);
    notmuch_status_t
    notmuch_message_add_property (notmuch_message_t *message,
                                  const char *key, const char *value);
    notmuch_status_t
    notmuch_message_remove_property (notmuch_message_t *message,
                                     const char *key, const char *value);
    notmuch_status_t
    notmuch_message_remove_all_properties (notmuch_message_t *message,
                                           const char *key);
    notmuch_message_properties_t *
    notmuch_message_get_properties (notmuch_message_t *message,
                                    const char *key, notmuch_bool_t exact);
    notmuch_bool_t
    notmuch_message_properties_valid (notmuch_message_properties_t
                                          *properties);
    void
    notmuch_message_properties_move_to_next (notmuch_message_properties_t
                                                 *properties);
    const char *
    notmuch_message_properties_key (notmuch_message_properties_t *properties);
    const char *
    notmuch_message_properties_value (notmuch_message_properties_t
                                          *properties);
    void
    notmuch_message_properties_destroy (notmuch_message_properties_t
                                            *properties);
    void
    notmuch_message_destroy (notmuch_message_t *message);

    notmuch_bool_t
    notmuch_tags_valid (notmuch_tags_t *tags);
    const char *
    notmuch_tags_get (notmuch_tags_t *tags);
    void
    notmuch_tags_move_to_next (notmuch_tags_t *tags);
    void
    notmuch_tags_destroy (notmuch_tags_t *tags);

    notmuch_bool_t
    notmuch_filenames_valid (notmuch_filenames_t *filenames);
    const char *
    notmuch_filenames_get (notmuch_filenames_t *filenames);
    void
    notmuch_filenames_move_to_next (notmuch_filenames_t *filenames);
    void
    notmuch_filenames_destroy (notmuch_filenames_t *filenames);
    notmuch_indexopts_t *
    notmuch_database_get_default_indexopts (notmuch_database_t *db);
    notmuch_status_t
    notmuch_indexopts_set_decrypt_policy (notmuch_indexopts_t *indexopts,
                                          notmuch_decryption_policy_t decrypt_policy);
    notmuch_decryption_policy_t
    notmuch_indexopts_get_decrypt_policy (const notmuch_indexopts_t *indexopts);
    void
    notmuch_indexopts_destroy (notmuch_indexopts_t *options);

    notmuch_status_t
    notmuch_database_set_config (notmuch_database_t *db, const char *key, const char *value);
    notmuch_status_t
    notmuch_database_get_config (notmuch_database_t *db, const char *key, char **value);
    notmuch_config_pairs_t *
    notmuch_config_get_pairs (notmuch_database_t *db, const char *prefix);
    notmuch_bool_t
    notmuch_config_pairs_valid (notmuch_config_pairs_t *config_list);
    const char *
    notmuch_config_pairs_key (notmuch_config_pairs_t *config_list);
    const char *
    notmuch_config_pairs_value (notmuch_config_pairs_t *config_list);
    void
    notmuch_config_pairs_move_to_next (notmuch_config_pairs_t *config_list);
    void
    notmuch_config_pairs_destroy (notmuch_config_pairs_t *config_list);
    """
)


if __name__ == '__main__':
    ffibuilder.compile(verbose=True)



# Generate all possible single-root message thread structures of size
# argv[1].  Each output line is a thread structure, where the n'th
# field is either a number giving the parent of message n or "None"
# for the root.
import sys
from itertools import chain, combinations

def subsets(s):
    return chain.from_iterable(combinations(s, r) for r in range(len(s)+1))

nodes = set(range(int(sys.argv[1])))

# Queue of (tree, free, to_expand) where tree is a {node: parent}
# dictionary, free is a set of unattached nodes, and to_expand is
# itself a queue of nodes in the tree that need to be expanded.
# The queue starts with all single-node trees.
queue = [({root: None}, nodes - {root}, (root,)) for root in nodes]

# Process queue
while queue:
    tree, free, to_expand = queue.pop()

    if len(to_expand) == 0:
        # Only print full-sized trees
        if len(free) == 0:
            print(" ".join(map(str, [msg[1] for msg in sorted(tree.items())])))
    else:
        # Expand node to_expand[0] with each possible set of children
        for children in subsets(free):
            ntree = {child: to_expand[0] for child in children}
            ntree.update(tree)
            nfree = free.difference(children)
            queue.append((ntree, nfree, to_expand[1:] + tuple(children)))



"""
This file is part of notmuch.

Notmuch is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

Notmuch is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with notmuch.  If not, see <https://www.gnu.org/licenses/>.

Copyright 2010 Sebastian Spaeth <Sebastian@SSpaeth.de>
"""

from ctypes import c_uint, c_long
from .globals import (
    nmlib,
    NotmuchDirectoryP,
    NotmuchFilenamesP
)
from .errors import (
    STATUS,
    NotmuchError,
    NotInitializedError,
)
from .filenames import Filenames

class Directory(object):
    """Represents a directory entry in the notmuch directory

    Modifying attributes of this object will modify the
    database, not the real directory attributes.

    The Directory object is usually derived from another object
    e.g. via :meth:`Database.get_directory`, and will automatically be
    become invalid whenever that parent is deleted. You should
    therefore initialized this object handing it a reference to the
    parent, preventing the parent from automatically being garbage
    collected.
    """

    """notmuch_directory_get_mtime"""
    _get_mtime = nmlib.notmuch_directory_get_mtime
    _get_mtime.argtypes = [NotmuchDirectoryP]
    _get_mtime.restype = c_long

    """notmuch_directory_set_mtime"""
    _set_mtime = nmlib.notmuch_directory_set_mtime
    _set_mtime.argtypes = [NotmuchDirectoryP, c_long]
    _set_mtime.restype = c_uint

    """notmuch_directory_get_child_files"""
    _get_child_files = nmlib.notmuch_directory_get_child_files
    _get_child_files.argtypes = [NotmuchDirectoryP]
    _get_child_files.restype = NotmuchFilenamesP

    """notmuch_directory_get_child_directories"""
    _get_child_directories = nmlib.notmuch_directory_get_child_directories
    _get_child_directories.argtypes = [NotmuchDirectoryP]
    _get_child_directories.restype = NotmuchFilenamesP

    def _assert_dir_is_initialized(self):
        """Raises a NotmuchError(:attr:`STATUS`.NOT_INITIALIZED)
        if dir_p is None"""
        if not self._dir_p:
            raise NotInitializedError()

    def __init__(self, path, dir_p, parent):
        """
        :param path:   The absolute path of the directory object.
        :param dir_p:  The pointer to an internal notmuch_directory_t object.
        :param parent: The object this Directory is derived from
                       (usually a :class:`Database`). We do not directly use
                       this, but store a reference to it as long as
                       this Directory object lives. This keeps the
                       parent object alive.
        """
        self._path = path
        self._dir_p = dir_p
        self._parent = parent

    def set_mtime(self, mtime):
        """Sets the mtime value of this directory in the database

        The intention is for the caller to use the mtime to allow efficient
        identification of new messages to be added to the database. The
        recommended usage is as follows:

        * Read the mtime of a directory from the filesystem

        * Call :meth:`Database.index_file` for all mail files in
          the directory

        * Call notmuch_directory_set_mtime with the mtime read from the
          filesystem.  Then, when wanting to check for updates to the
          directory in the future, the client can call :meth:`get_mtime`
          and know that it only needs to add files if the mtime of the
          directory and files are newer than the stored timestamp.

          .. note::

                :meth:`get_mtime` function does not allow the caller to
                distinguish a timestamp of 0 from a non-existent timestamp. So
                don't store a timestamp of 0 unless you are comfortable with
                that.

        :param mtime: A (time_t) timestamp
        :raises: :exc:`XapianError` a Xapian exception occurred, mtime
                 not stored
        :raises: :exc:`ReadOnlyDatabaseError` the database was opened
                 in read-only mode so directory mtime cannot be modified
        :raises: :exc:`NotInitializedError` the directory object has not
                 been initialized
        """
        self._assert_dir_is_initialized()
        status = Directory._set_mtime(self._dir_p, mtime)

        if status != STATUS.SUCCESS:
            raise NotmuchError(status)

    def get_mtime(self):
        """Gets the mtime value of this directory in the database

        Retrieves a previously stored mtime for this directory.

        :param mtime: A (time_t) timestamp
        :raises: :exc:`NotmuchError`:

                        :attr:`STATUS`.NOT_INITIALIZED
                          The directory has not been initialized
        """
        self._assert_dir_is_initialized()
        return Directory._get_mtime(self._dir_p)

    # Make mtime attribute a property of Directory()
    mtime = property(get_mtime, set_mtime, doc="""Property that allows getting
                     and setting of the Directory *mtime* (read-write)

                     See :meth:`get_mtime` and :meth:`set_mtime` for usage and
                     possible exceptions.""")

    def get_child_files(self):
        """Gets a Filenames iterator listing all the filenames of
        messages in the database within the given directory.

        The returned filenames will be the basename-entries only (not
        complete paths.
        """
        self._assert_dir_is_initialized()
        files_p = Directory._get_child_files(self._dir_p)
        return Filenames(files_p, self)

    def get_child_directories(self):
        """Gets a :class:`Filenames` iterator listing all the filenames of
        sub-directories in the database within the given directory

        The returned filenames will be the basename-entries only (not
        complete paths.
        """
        self._assert_dir_is_initialized()
        files_p = Directory._get_child_directories(self._dir_p)
        return Filenames(files_p, self)

    @property
    def path(self):
        """Returns the absolute path of this Directory (read-only)"""
        return self._path

    def __repr__(self):
        """Object representation"""
        return "<notmuch Directory object '%s'>" % self._path

    _destroy = nmlib.notmuch_directory_destroy
    _destroy.argtypes = [NotmuchDirectoryP]
    _destroy.restype = None

    def __del__(self):
        """Close and free the Directory"""
        if self._dir_p:
            self._destroy(self._dir_p)



"""Tests for the behaviour of immutable and mutable tagsets.

This module tests the Pythonic behaviour of the sets.
"""

import collections
import subprocess
import textwrap

import pytest

from notmuch2 import _database as database
from notmuch2 import _tags as tags


class TestImmutable:

    @pytest.fixture
    def tagset(self, maildir, notmuch):
        """An non-empty immutable tagset.

        This will have the default new mail tags: inbox, unread.
        """
        maildir.deliver()
        notmuch('new')
        with database.Database(maildir.path, config=database.Database.CONFIG.EMPTY) as db:
            yield db.tags

    def test_type(self, tagset):
        assert isinstance(tagset, tags.ImmutableTagSet)
        assert isinstance(tagset, collections.abc.Set)

    def test_hash(self, tagset, maildir, notmuch):
        h0 = hash(tagset)
        notmuch('tag', '+foo', '*')
        with database.Database(maildir.path, config=database.Database.CONFIG.EMPTY) as db:
            h1 = hash(db.tags)
        assert h0 != h1

    def test_eq(self, tagset):
        assert tagset == tagset

    def test_neq(self, tagset, maildir, notmuch):
        notmuch('tag', '+foo', '*')
        with database.Database(maildir.path, config=database.Database.CONFIG.EMPTY) as db:
            assert tagset != db.tags

    def test_contains(self, tagset):
        print(tuple(tagset))
        assert 'unread' in tagset
        assert 'foo' not in tagset

    def test_isdisjoint(self, tagset):
        assert tagset.isdisjoint(set(['spam', 'ham']))
        assert not tagset.isdisjoint(set(['inbox']))

    def test_issubset(self, tagset):
        assert {'inbox'} <= tagset
        assert {'inbox'}.issubset(tagset)
        assert tagset <= {'inbox', 'unread', 'spam'}
        assert tagset.issubset({'inbox', 'unread', 'spam'})

    def test_issuperset(self, tagset):
        assert {'inbox', 'unread', 'spam'} >= tagset
        assert {'inbox', 'unread', 'spam'}.issuperset(tagset)
        assert tagset >= {'inbox'}
        assert tagset.issuperset({'inbox'})

    def test_iter(self, tagset):
        expected = sorted(['unread', 'inbox'])
        found = []
        for tag in tagset:
            assert isinstance(tag, str)
            found.append(tag)
        assert expected == sorted(found)

    def test_special_iter(self, tagset):
        expected = sorted([b'unread', b'inbox'])
        found = []
        for tag in tagset.iter():
            assert isinstance(tag, bytes)
            found.append(tag)
        assert expected == sorted(found)

    def test_special_iter_codec(self, tagset):
        for tag in tagset.iter(encoding='ascii', errors='surrogateescape'):
            assert isinstance(tag, str)

    def test_len(self, tagset):
        assert len(tagset) == 2

    def test_and(self, tagset):
        common = tagset & {'unread'}
        assert isinstance(common, set)
        assert isinstance(common, collections.abc.Set)
        assert common == {'unread'}
        common = tagset.intersection({'unread'})
        assert isinstance(common, set)
        assert isinstance(common, collections.abc.Set)
        assert common == {'unread'}

    def test_or(self, tagset):
        res = tagset | {'foo'}
        assert isinstance(res, set)
        assert isinstance(res, collections.abc.Set)
        assert res == {'unread', 'inbox', 'foo'}
        res = tagset.union({'foo'})
        assert isinstance(res, set)
        assert isinstance(res, collections.abc.Set)
        assert res == {'unread', 'inbox', 'foo'}

    def test_sub(self, tagset):
        res = tagset - {'unread'}
        assert isinstance(res, set)
        assert isinstance(res, collections.abc.Set)
        assert res == {'inbox'}
        res = tagset.difference({'unread'})
        assert isinstance(res, set)
        assert isinstance(res, collections.abc.Set)
        assert res == {'inbox'}

    def test_rsub(self, tagset):
        res = {'foo', 'unread'} - tagset
        assert isinstance(res, set)
        assert isinstance(res, collections.abc.Set)
        assert res == {'foo'}

    def test_xor(self, tagset):
        res = tagset ^ {'unread', 'foo'}
        assert isinstance(res, set)
        assert isinstance(res, collections.abc.Set)
        assert res == {'inbox', 'foo'}
        res = tagset.symmetric_difference({'unread', 'foo'})
        assert isinstance(res, set)
        assert isinstance(res, collections.abc.Set)
        assert res == {'inbox', 'foo'}

    def test_rxor(self, tagset):
        res = {'unread', 'foo'} ^ tagset
        assert isinstance(res, set)
        assert isinstance(res, collections.abc.Set)
        assert res == {'inbox', 'foo'}

    def test_copy(self, tagset):
        res = tagset.copy()
        assert isinstance(res, set)
        assert isinstance(res, collections.abc.Set)
        assert res == {'inbox', 'unread'}


class TestMutableTagset:

    @pytest.fixture
    def tagset(self, maildir, notmuch):
        """An non-empty mutable tagset.

        This will have the default new mail tags: inbox, unread.
        """
        _, pathname = maildir.deliver()
        notmuch('new')
        with database.Database(maildir.path,
                               mode=database.Mode.READ_WRITE,
                               config=database.Database.CONFIG.EMPTY) as db:
            msg = db.get(pathname)
            yield msg.tags

    def test_type(self, tagset):
        assert isinstance(tagset, collections.abc.MutableSet)
        assert isinstance(tagset, tags.MutableTagSet)

    def test_hash(self, tagset):
        assert not isinstance(tagset, collections.abc.Hashable)
        with pytest.raises(TypeError):
            hash(tagset)

    def test_add(self, tagset):
        assert 'foo' not in tagset
        tagset.add('foo')
        assert 'foo' in tagset

    def test_discard(self, tagset):
        assert 'inbox' in tagset
        tagset.discard('inbox')
        assert 'inbox' not in tagset

    def test_discard_not_present(self, tagset):
        assert 'foo' not in tagset
        tagset.discard('foo')

    def test_clear(self, tagset):
        assert len(tagset) > 0
        tagset.clear()
        assert len(tagset) == 0

    def test_from_maildir_flags(self, maildir, notmuch):
        _, pathname = maildir.deliver(flagged=True)
        notmuch('new')
        with database.Database(maildir.path,
                               mode=database.Mode.READ_WRITE,
                               config=database.Database.CONFIG.EMPTY) as db:
            msg = db.get(pathname)
            msg.tags.discard('flagged')
            msg.tags.from_maildir_flags()
            assert 'flagged' in msg.tags

    def test_to_maildir_flags(self, maildir, notmuch):
        _, pathname = maildir.deliver(flagged=True)
        notmuch('new')
        with database.Database(maildir.path,
                               mode=database.Mode.READ_WRITE,
                               config=database.Database.CONFIG.EMPTY) as db:
            msg = db.get(pathname)
            flags = msg.path.name.split(',')[-1]
            assert 'F' in flags
            msg.tags.discard('flagged')
            msg.tags.to_maildir_flags()
            flags = msg.path.name.split(',')[-1]
            assert 'F' not in flags

    def test_isdisjoint(self, tagset):
        assert tagset.isdisjoint(set(['spam', 'ham']))
        assert not tagset.isdisjoint(set(['inbox']))

    def test_issubset(self, tagset):
        assert {'inbox'} <= tagset
        assert {'inbox'}.issubset(tagset)
        assert not {'spam'} <= tagset
        assert not {'spam'}.issubset(tagset)
        assert tagset <= {'inbox', 'unread', 'spam'}
        assert tagset.issubset({'inbox', 'unread', 'spam'})
        assert not {'inbox', 'unread', 'spam'} <= tagset
        assert not {'inbox', 'unread', 'spam'}.issubset(tagset)

    def test_issuperset(self, tagset):
        assert {'inbox', 'unread', 'spam'} >= tagset
        assert {'inbox', 'unread', 'spam'}.issuperset(tagset)
        assert tagset >= {'inbox'}
        assert tagset.issuperset({'inbox'})

    def test_union(self, tagset):
        assert {'spam'}.union(tagset) == {'inbox', 'unread', 'spam'}
        assert tagset.union({'spam'}) == {'inbox', 'unread', 'spam'}



# -*- coding: utf-8 -*-

import os.path

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'notmuch'
authors = 'Carl Worth and many others'
copyright = '2009-2015, {0}'.format(authors)

location = os.path.dirname(__file__)

dirname = location
while True:
    version_file = os.path.join(dirname, 'version')
    if os.path.exists(version_file):
        with open(version_file,'r') as f:
            version = f.read().strip()
            break
    if dirname == '/':
        raise ValueError(
            'no version file found in this directory or its ancestors')
    dirname = os.path.dirname(dirname)

# The full version, including alpha/beta/rc tags.
release = version

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).

man_pages = [
    ('man1/notmuch-report.1', 'notmuch-report',
     'generate reports from notmuch queries', [authors], 1),
    ('man5/notmuch-report.json.5', 'notmuch-report.json',
     'configure notmuch-report', [authors], 5),
]

# If true, show URL addresses after external links.
#man_show_urls = False

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
# If true, do not generate a @detailmenu in the "Top" node's menu.
texinfo_no_detailmenu = True

texinfo_documents = [
    ('man1/notmuch-report.1', 'notmuch-report',
     'generate reports from notmuch queries', authors, 'notmuch-report',
     'generate reports from notmuch queries', 'Miscellaneous'),
    ('man5/notmuch-report.json.5', 'notmuch-report.json',
     'configure notmuch-report', authors, 'notmuch-report.json',
     'configure notmuch-report', 'Miscellaneous'),
]
